# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

from hashlib import md5

import sympy as sp

from .utils import align, concat_generator_results
from .wrapper import WrapperBase


class InterpolationFunction1DInstance(
    sp.Function("_InterpolationFunction1D", nargs=2, real=True)
):
    @property
    def free_symbols(self):
        return self.argument.free_symbols

    @property
    def _name(self):
        return self.args[0]

    @property
    def argument(self):
        return self.args[1]

    def __hash__(self):
        return hash((self._name, self.argument))

    def __eq__(self, other):
        if type(self) != type(other):  # noqa: E721
            return False
        return self._name == other._name and self.argument == other.argument

    def __str__(self):
        return "{}({})".format(self._name, self.argument)

    def _subs(self, old, new, **hints):
        return InterpolationFunction1DInstance(
            self.args[0], self.args[1]._subs(old, new, **hints)
        )


class InterpolationFunction1D(
    sp.Function("_InterpolationFunction1D", nargs=1, real=True)
):
    @property
    def free_symbols(self):
        return {self}

    @property
    def _name(self):
        return self.args[0]

    def __hash__(self):
        return hash((self._name,))

    def __eq__(self, other):
        if type(self) != type(other):  # noqa: E721
            return False
        return self._name == other._name

    def __call__(self, argument):
        return InterpolationFunction1DInstance(self._name, argument)

    def get_unique_id(self):
        return md5(str(self._name).encode("utf-8")).hexdigest()

    def _subs(self, old, new, **hints):
        return InterpolationFunction1D(self.args[0]._subs(old, new, **hints))


class InterpolationFunction1DWrapper(WrapperBase):
    def __init__(self, interpolation_function_1d, visitor):
        # self.namejkjjk = interpolation_function_1d
        self.name = interpolation_function_1d._name
        self.visitor = visitor

    def setup_code_generation(self):
        pass

    def get_unique_id(self):
        return md5(str(self.name).encode("utf-8")).hexdigest()

    def c_header(self):
        return align(
            """
            |extern "C" {{
            |   void         _set_{name}_values(double *, double *, size_t);
            |}}
            |
            |extern "C" double __ip_{name}(double);
            |extern "C" void   __ip_sorted_{name}(double *, size_t, double *);
            |
            |extern "C" double  * _{name}_x_min_p, * _{name}_x_max_p;
            |extern "C" double  * _get_x_{name}();
            |extern "C" double  * _get_y_{name}();
            |extern "C" int _get_n_{name}();
            """.format(name=self.name)
        )

    def c_code(self, header_file_path):
        return align(
            """
            |gsl_interp_accel * {name}_interp_accel = 0;
            |gsl_spline       * {name}_spline = 0;
            |static double    _{name}_x_min, _{name}_x_max;
            |double           * _{name}_x_min_p = NULL, * _{name}_x_max_p = NULL;
            |static double    *_{name}_x = NULL, *_{name}_y;
            |static int       _{name}_n;
            |
            |extern "C" double * _get_x_{name}() {{
            |    return _{name}_x;
            |}}
            |
            |extern "C" double * _get_y_{name}() {{
            |    return _{name}_y;
            |}}
            |
            |extern "C" int _get_n_{name}() {{
            |    return _{name}_n;
            |}}
            |
            |extern "C" void _set_{name}_values(double *x, double *y, size_t n) {{
            |     if ({name}_interp_accel) {{
            |         gsl_spline_free({name}_spline);
            |         gsl_interp_accel_free({name}_interp_accel);
            |     }}
            |     {name}_interp_accel = gsl_interp_accel_alloc();
            |     if ({name}_interp_accel == 0) {{
            |          set_error_message_gsl("_set_{name}_values", GSL_ENOMEM);
            |          return;
            |     }}
            |     {name}_spline = gsl_spline_alloc(gsl_interp_cspline, n);
            |     if ({name}_spline == 0) {{
            |          set_error_message_gsl("_set_{name}_values", GSL_ENOMEM);
            |          return;
            |     }}
            |     gsl_spline_init({name}_spline, x, y, n);
            |
            |
            |     if (_{name}_x != NULL) {{
            |        free(_{name}_x);
            |        free(_{name}_y);
            |     }}
            |     _{name}_x = (double *)malloc(n * sizeof(double));
            |     _{name}_y = (double *)malloc(n * sizeof(double));
            |
            |     /* setting pointers to value different to NULL indicactes that
            |        this function was called: */
            |     _{name}_x_min = _{name}_x_max = x[0];
            |     _{name}_x[0] = x[0];
            |     _{name}_y[0] = y[0];
            |
            |     for (size_t i = 1; i < n; ++i)
            |     {{
            |         if (x[i] > _{name}_x_max) _{name}_x_max = x[i];
            |         if (x[i] < _{name}_x_min) _{name}_x_min = x[i];
            |         _{name}_x[i] = x[i];
            |         _{name}_y[i] = y[i];
            |     }}
            |     _{name}_n = n;
            |
            |     _{name}_x_min_p = & _{name}_x_min;
            |     _{name}_x_max_p = & _{name}_x_max;
            |
            |}}
            |
            |extern "C" double __ip_{name}(double x) {{
            |     double result = 0.0;
            |     int    error_code = 0;
            |     if ({name}_spline) {{
            |         if (x > _{name}_x_max) x = _{name}_x_max;
            |         error_code = gsl_spline_eval_e({name}_spline, x,
            |                                        {name}_interp_accel, &result);
            |         if (error_code)
            |              set_error_message_gsl("spline eval failed", error_code);
            |     }}
            |     else
            |         set_error_message("you must call set_{name}_values first");
            |     return result;
            |}}
            |
            |extern "C" void __ip_sorted_{name}(double *x, size_t n, double* y) {{
            |     if ({name}_spline) {{
            |         cspline_eval_vector_sorted({name}_spline->interp->state,
            |                                    {name}_spline->x,
            |                                    {name}_spline->y,
            |                                    {name}_spline->interp->size,
            |                                    x, n, y);
            |     }}
            |     else
            |         set_error_message("you must call set_{name}_values first");
            |}}
            """.format(name=self.name)
        )

    @concat_generator_results
    def cython_code(self, header_file_path):
        # we insert "_" at beginning of c functions to avoid clash with
        # cython functions:
        for code in self.cython_header(header_file_path):
            yield code
        for code in self.cython_function_wrapper():
            yield code
        return
        for code in self.cython_ufunc_wrapper():
            yield code

    def cython_header(self, header_file_path):
        yield align(
            """
        |cdef extern from "{header_file_path}":
        |    double __ip_{name}(double);
        |    void   __ip_sorted_{name}(double *, size_t, double *);
        |    double _set_{name}_values(double *, double *, size_t);
        |    double  * _{name}_x_min_p, * _{name}_x_max_p;
        |    double  * _get_x_{name}();
        |    double  * _get_y_{name}();
        |    int _get_n_{name}();
        """.lstrip().format(header_file_path=header_file_path, name=self.name)
        )

    def cython_function_wrapper(self):
        yield align(
            """
        |def set_{name}_values(np.ndarray[np.double_t, ndim=1, cast=True, mode="c"] x,
        |                      np.ndarray[np.double_t, ndim=1, cast=True, mode="c"] y):
        |
        |    assert len(x) == len(y), "x and y must be of same size"
        |    assert len(x) > 0, "empty array not allowed"
        |    _set_{name}_values(&x[0], &y[0], len(x))
        |
        |def ip_sorted_{name}(np.ndarray[np.double_t, ndim=1, cast=True] x):
        |    cdef result = np.zeros_like(x, np.float64)
        |    cdef double[:] _result = result
        |    cdef double[:] _x = x.ravel()
        |    __ip_sorted_{name}(&_x[0], x.shape[0], &_result[0])
        |    return result
        |
        |def get_x_{name}():
        |    if _get_x_{name}() == NULL:
        |       return None
        |    return np.asarray(<np.float64_t[:_get_n_{name}()]> _get_x_{name}())
        |
        |def get_y_{name}():
        |    if _get_x_{name}() == NULL:
        |       return None
        |    return np.asarray(<np.float64_t[:_get_n_{name}()]> _get_y_{name}())
        |
        """
        ).format(name=self.name)

    def determine_required_extra_wrappers(self):
        pass



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess

from setuptools._distutils import ccompiler

from .c_code import LSODA_PATH
from .utils import base_cache_folder, create_folder_if_not_exists, get_platform

CC = os.environ.get("CC", ccompiler.new_compiler().executables["compiler"][0])
HERE = os.path.dirname(os.path.abspath(__file__))


def compile_if_needed(target_folder):
    lsoda_out = os.path.join(target_folder, "lsoda.o")
    if not os.path.exists(lsoda_out):
        output = subprocess.check_output(
            [CC, "-c", LSODA_PATH, "-fPIC", "-o", lsoda_out],
            shell=False,
            universal_newlines=True,
        )
        print(output)
    return lsoda_out


def install_lsoda_if_needed(lsoda_root_folder=None):
    if lsoda_root_folder is None:
        lsoda_root_folder = os.path.join(base_cache_folder(), "sympy2c", get_platform())

    target_folder = os.path.join(lsoda_root_folder, "lsoda")
    create_folder_if_not_exists(target_folder)
    object_file = compile_if_needed(target_folder)
    return object_file


if __name__ == "__main__":
    print(install_lsoda_if_needed())



# This file is part of PyCosmo, a multipurpose cosmology calculation tool in Python.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.


import pickle

import numpy as np
import sympy as sp

from sympy2c import (
    Function,
    Globals,
    Integral,
    InterpolationFunction1D,
    Module,
    Symbol,
    Vector,
    symbols,
)
from sympy2c.ode import Ode
from sympy2c.ode_fast import OdeFast


def test_lsoda_solver(tmpdir, compile_and_load):
    v1x, v2x, v1y, v2y = symbols("v1x v2x v1y v2y")
    p1x, p2x, p1y, p2y = symbols("p1x p2x p1y p2y")

    t, m1, m2 = symbols("t m1 m2")

    p1x_dot = v1x
    p1y_dot = v1y

    p2x_dot = v2x
    p2y_dot = v2y

    d2 = ((p1x - p2x) ** 2 + (p1y - p2y) ** 2) ** 0.5

    v1x_dot = -m2 * (p1x - p2x) / d2**3
    v1y_dot = -m2 * (p1y - p2y) / d2**3

    v2x_dot = -m1 * (p2x - p1x) / d2**3
    v2y_dot = -m1 * (p2y - p1y) / d2**3

    ode_0 = OdeFast(
        "planets_0",
        t,
        [p1x, p1y, p2x, p2y, v1x, v1y, v2x, v2y],
        [p1x_dot, p1y_dot, p2x_dot, p2y_dot, v1x_dot, v1y_dot, v2x_dot, v2y_dot],
    )

    # we crate same solver again to check if overlap creates compilation issues:
    ode_1 = OdeFast(
        "planets_1",
        t,
        [p1x, p1y, p2x, p2y, v1x, v1y, v2x, v2y],
        [p1x_dot, p1y_dot, p2x_dot, p2y_dot, v1x_dot, v1y_dot, v2x_dot, v2y_dot],
    )

    ode_2 = OdeFast(
        "planets_2",
        t,
        [p1x, p1y, p2x, p2y, v1x, v1y, v2x, v2y],
        [p1x_dot, p1y_dot, p2x_dot, p2y_dot, v1x_dot, v1y_dot, v2x_dot, v2y_dot],
        splits=[4, 6],
    )

    # gamma variable conflicts with C gamma function, thus we use gamma_ instead:
    x, y, alpha, beta, gamma, delta = symbols("x y alpha beta gamma_ delta")

    ode_3 = OdeFast(
        "predator_prey",
        t,
        [x, y],
        [alpha * x - beta * x * y, delta * x * y - gamma * y],
    )

    module = Module()
    module.add(Globals(m1, m2, alpha, beta, gamma, delta))
    module.add(ode_0)
    module.add(ode_1)
    module.add(ode_2)
    module.add(ode_3)

    assert module.get_unique_id() is not None

    w = compile_and_load(module)
    assert w.get_unique_id() is not None

    w.globals.m1 = 2
    w.globals.m2 = 1

    w = pickle.loads(pickle.dumps(w))

    p1 = (5, 0)
    v1 = (0.2, -0.2)

    p2 = (10, 0)
    v2 = (0, 0.5)

    t = np.linspace(0, 60, 1000)

    tobe = [
        10.419238664939208,
        1.4668800372218,
        11.161522670121593,
        0.06623992555639961,
        2.6002221831463435,
        1.3737285597860913,
        2.8858456145722458,
        2.017801931489659,
    ]

    for solver in (
        w.solve_fast_planets_0,
        w.solve_fast_planets_1,
        w.solve_fast_planets_2,
    ):
        y = solver(
            np.array(
                (p1[0], p1[1], p2[0], p2[1], v1[0], v1[1], v2[0], v2[1]), dtype=float
            ),
            t,
            1e-5,
            1e-5 * np.ones((8,)),
        )

        result, meta = y

        assert sum(meta["steps"]) == 100
        assert abs(sum(meta["step_sizes"]) - t[-1] + t[0]) < 1e-8

        assert meta["istate"] == 2
        assert meta["new_traces"] == {}

        p1x, p1y, p2x, p2y, v1x, v1y, v2x, v2y = result.T

        is_ = [
            p1x.mean(),
            p1y.mean(),
            p2x.mean(),
            p2y.mean(),
            p1x.std(),
            p1y.std(),
            p2x.std(),
            p2y.std(),
        ]

        rel_diffs = [abs(i - t) / abs(t) for (i, t) in zip(is_, tobe)]
        assert all([rel_diff < 1e-8 for rel_diff in rel_diffs])

    w.globals.alpha = 2 / 3
    w.globals.beta = 4 / 3
    w.globals.gamma_ = 1
    w.globals.delta = 1

    t = np.linspace(0, 10, 10)

    result, meta = w.solve_fast_predator_prey(
        np.array((1.0, 0.2)), t, 1e-5, 1e-5 * np.ones((2,))
    )

    tobe = np.array(
        [
            [1.0, 0.2],
            [1.5147443, 0.2638374],
            [1.77945534, 0.57985881],
            [1.09852698, 1.00260696],
            [0.57480453, 0.79116964],
            [0.48529783, 0.45766939],
            [0.6036409, 0.27152023],
            [0.90179806, 0.20262532],
            [1.38857545, 0.23467216],
            [1.79452756, 0.47218948],
        ]
    )

    assert np.allclose(result, tobe)

    assert np.allclose(
        w.planets_0_dot(0.0, np.arange(8.0)),
        np.array([4.0, 5.0, 6.0, 7.0, 0.08838835, 0.08838835, -0.1767767, -0.1767767]),
    )
    assert sum(meta["steps"]) > 0


def test_split(compile_and_load):
    t, y0, y1, y2, y3, y4, y5, y6, alpha = symbols("t y0 y1 y2 y3 y4 y5 y6 alpha")

    ydot0 = -0.04 * y0 + 1e4 * y1 * y2
    ydot2 = 3e7 * y1 * y1
    ydot1 = -ydot0 - ydot2

    ydot3 = y4
    ydot4 = y5
    ydot5 = -(alpha**3) * y3

    ode_baseline = Ode(
        "robertson_baseline",
        t,
        [y0, y1, y2, y3, y4, y5],
        [ydot0, ydot1, ydot2, ydot3, ydot4, ydot5],
    )

    ode_0 = OdeFast(
        "robertson_0",
        t,
        [y0, y1, y2, y3, y4, y5],
        [ydot0, ydot1, ydot2, ydot3, ydot4, ydot5],
        splits=[1],  # can trigger issue with rec_inv related compuations
    )
    ode_1 = OdeFast(
        "robertson_1",
        t,
        [y0, y1, y2, y3, y4, y5],
        [ydot0, ydot1, ydot2, ydot3, ydot4, ydot5],
        splits=[2, 5],  # trigger multiple splits
    )
    ode_2 = OdeFast(
        "robertson_2",
        t,
        [y0, y1, y2, y3, y4, y5],
        [ydot0, ydot1, ydot2, ydot3, ydot4, ydot5],
    )

    module = Module()
    module.add(ode_baseline)
    module.add(ode_0)
    module.add(ode_1)
    module.add(ode_2)
    module.add(Globals(alpha))

    wrapper = compile_and_load(module)

    tvec = 0.4 * 10 ** np.arange(0, 6)
    rtol = 1e-6
    atol = np.array([1e-9, 1e-8, 1e-10, 1e-8])
    atol = 1e-8

    alpha = 1e-4
    wrapper.globals.alpha = alpha
    y0 = np.array([1.0, 0.0, 0.0, 1.0, -alpha, alpha**2])

    result_baseline, meta_baseline = wrapper.solve_robertson_baseline(
        y0, tvec, rtol, atol, use_jacobian=True
    )
    print(meta_baseline)

    result_0, meta_0 = wrapper.solve_fast_robertson_0(y0, tvec, rtol, atol)
    print(meta_0)

    result_1, meta_1 = wrapper.solve_fast_robertson_1(y0, tvec, rtol, atol)
    print(meta_1)

    result_2, meta_2 = wrapper.solve_fast_robertson_2(y0, tvec, rtol, atol)
    print(meta_2)

    result_3, meta_3 = wrapper.solve_fast_robertson_0(
        y0, tvec, rtol, atol, enable_fast_solver=0
    )
    print(meta_3)

    assert (
        np.max(np.abs(result_baseline[1:] - result_0[1:]) / result_baseline[1:]) < 1e-4
    )
    assert (
        np.max(np.abs(result_baseline[1:] - result_1[1:]) / result_baseline[1:]) < 1e-4
    )
    assert (
        np.max(np.abs(result_baseline[1:] - result_2[1:]) / result_baseline[1:]) < 1e-4
    )
    assert (
        np.max(np.abs(result_baseline[1:] - result_3[1:]) / result_baseline[1:]) < 1e-4
    )


def test_compare_lsoda(tmpdir):
    """example problem from lsoda.f source code"""

    t, y1, y2, y3 = symbols("t y1 y2 y3")

    ydot1 = -0.04 * y1 + 1e4 * y2 * y3
    ydot3 = 3e7 * y2 * y2
    ydot2 = -ydot1 - ydot3

    # we change order of original system of odes to enforce permutations during modified
    # lsoda solver:
    ode_0 = OdeFast(
        "robertson_split", t, [y3, y2, y1], [ydot3, ydot2, ydot1], splits=[1]
    )
    ode_1 = OdeFast("robertson_nonsplit", t, [y3, y2, y1], [ydot3, ydot2, ydot1])
    ode_2 = Ode("robertson", t, [y3, y2, y1], [ydot3, ydot2, ydot1])

    module = Module()
    module.add(ode_0)
    module.add(ode_1)
    module.add(ode_2)

    tvec = 0.4 * 10 ** np.arange(0, 6)
    rtol = 1e-6
    atol = np.array([1e-8, 1e-10, 1e-8])

    root_folder = tmpdir.join("root_folder").strpath
    lsoda_folder = tmpdir.join("lsoda").strpath

    wrapper = module.compile_and_load(
        root_folder=root_folder, lsoda_folder=lsoda_folder
    )
    wrapper.set_sec_factor(1.0)

    y0 = np.array([0, 0, 1.0])
    result_fast_split, meta = wrapper.solve_fast_robertson_split(y0, tvec, rtol, atol)
    assert meta["new_traces"] == {}
    assert sum(meta["steps"]) > 0
    assert meta["istate"] > 0, meta

    result_fast_nonsplit, meta = wrapper.solve_fast_robertson_nonsplit(
        y0, tvec, rtol, atol
    )
    assert meta["new_traces"] == {0: [[1, 1, 2]]}
    assert sum(meta["steps"]) > 0
    assert meta["istate"] > 0, meta

    result, _ = wrapper.solve_robertson(y0, tvec, rtol, atol, use_jacobian=True)

    assert np.allclose(result_fast_split, result)
    assert np.allclose(result_fast_nonsplit, result)
    assert np.allclose(result_fast_nonsplit, result_fast_split)

    wrapper_path_before = wrapper.__file__

    wrapper = module.recompile_and_load(
        root_folder=root_folder, lsoda_folder=lsoda_folder
    )
    assert wrapper_path_before != wrapper.__file__

    result_fast_nonsplit, meta = wrapper.solve_fast_robertson_nonsplit(
        y0, tvec, rtol, atol
    )
    assert meta["new_traces"] == {}
    assert sum(meta["steps"]) > 0
    assert meta["istate"] > 0, meta

    assert np.allclose(result_fast_nonsplit, result)
    assert np.allclose(result_fast_nonsplit, result_fast_split)

    wrapper_path_before = wrapper.__file__
    w2 = module.recompile_and_load(root_folder=root_folder, lsoda_folder=lsoda_folder)
    assert wrapper_path_before == w2.__file__


def test_with_time_var_in_rhs(compile_and_load):
    x, t, a = symbols("x t a")

    ode_0 = OdeFast("test_0", t, [x], [a * t / x])
    ode_1 = Ode("test_1", t, [x], [a * t / x])
    module = Module()
    module.add(Globals(a))
    module.add(ode_0)
    module.add(ode_1)

    wrapper = compile_and_load(module)

    tvec = np.linspace(1, 20, 21)
    a = 1
    y0 = np.array((20.0,))
    wrapper.globals.a = a
    result_1, meta_1 = wrapper.solve_test_1(y0, tvec, 1e-4, 1e-4)

    assert meta_1["istate"] > 0

    result_0, meta_0 = wrapper.solve_fast_test_0(y0, tvec, 1e-4, 1e-4)
    assert meta_0["istate"] > 0

    assert meta_0["new_traces"] == {}
    assert sum(meta_0["steps"]) > 0

    assert np.allclose(result_1, result_0)


def test_vector_symbols(compile_and_load):
    state = Vector("s", 2)
    t = Symbol("t")

    # solution of this ode are sums of sin and cos, depending on the starting values
    lhs = [state[0], state[1]]
    rhs = [state[1], -state[0]]

    ode = OdeFast("wave", t, lhs, rhs)
    module = Module()
    module.add(ode)

    wrapper = compile_and_load(module)
    assert wrapper is not None

    t = np.linspace(0, 6, 10)

    result, meta = wrapper.solve_fast_wave(np.array([0.0, 1.0]), t, 1e-10, 1e-10)
    assert meta["new_traces"] == {}
    assert sum(meta["steps"]) > 0

    assert np.allclose(result[:, 0], np.sin(t))
    assert np.allclose(result[:, 1], np.cos(t))


def test_with_integral_and_interpolation(compile_and_load):
    x, tau, t = symbols("x tau t")
    squared = InterpolationFunction1D("squared")
    exp_as_integral = Integral(sp.exp(tau), tau, 0, t) + 1

    lhs = symbols("y1 y2")
    rhs = [lhs[1] / exp_as_integral, lhs[0] / squared(t)]
    ode = OdeFast("mixed", t, lhs, rhs)

    module = Module()
    module.add(ode)

    wrapper = compile_and_load(module)

    tvec = np.linspace(0, 4, 100)

    wrapper.set_squared_values(tvec, tvec**2)

    module2 = Module()
    module2.add(ode)
    module2.add(Function("squared", squared(x), x))
    module2.add(Function("exp", exp_as_integral, t))

    compile_and_load(module2)


def test_with_integral_with_cse_issue(compile_and_load):
    q = Symbol("q")
    t = Symbol("t")

    integrand = q**2 + q**4 + q**6
    ii = Integral(integrand * integrand, q, 0, 1)
    jj = Integral(integrand * q, q, 0, 1)

    lhs = (u, v) = symbols("u v")
    rhs = [ii * u + v, jj * v + u]

    f = Function("f", [ii, jj])
    g = Function("g", [ii, jj])

    ode = OdeFast("mixed", t, lhs, rhs)

    module = Module()
    module.add(ode)
    module.add(f)
    module.add(g)

    compile_and_load(module)


def test_highdim_ode(tmpdir):
    state = list(Vector("y", 100))
    (t,) = symbols("t")
    rhs = [-yi + yin for yi, yin in zip(state, state[1:] + state[:1])]
    ode = OdeFast("large", t, state, rhs)

    module = Module()
    module.add(ode)

    root_folder = tmpdir.join("root_folder").strpath
    lsoda_folder = tmpdir.join("lsoda").strpath

    wrapper = module.compile_and_load(
        root_folder=root_folder, lsoda_folder=lsoda_folder
    )

    y0 = np.arange(100) / 99.0
    tvec = np.linspace(0, 5, 100)

    solution, meta = wrapper.solve_fast_large(y0, tvec, atol=1e-5, rtol=1e-5)
    assert meta["istate"] == 2


def test_reordering_rows(tmpdir):
    t, y0, y1, y2, y3, y4, y5 = symbols("t y0 y1 y2 y3 y4 y5")

    lhs = [y0, y1, y2]

    ydot0 = -0.04 * y0 + 1e4 * y1 * y2
    ydot2 = 3e7 * y1 * y1
    ydot1 = -ydot0 - ydot2

    rhs = [ydot0, ydot1, ydot2]

    y0_vec = [1.0, 0.0, 0.0]
    atol = [1e-8, 1e-8, 1e-10]
    rtol = 1e-6

    ode_robertson = OdeFast("robertson", t, lhs, rhs)

    lhs_extended = [y0, y1, y2, y3, y4, y5][::-1]

    ydot3 = -0.04 * y3 + 1e4 * y4 * y5
    ydot5 = 3e7 * y4 * y4
    ydot4 = -ydot3 - ydot5

    rhs_extended = [ydot0, ydot1, ydot2, ydot3, ydot4, ydot5][::-1]
    atol_extended = (atol + atol)[::-1]
    y0_vec_extended = (y0_vec + y0_vec)[::-1]

    ode_baseline = OdeFast("baseline", t, lhs_extended, rhs_extended)
    ode_reordered = OdeFast("reordered", t, lhs_extended, rhs_extended, reorder=True)
    ode_reordered_with_splits = OdeFast(
        "reordered_with_splits", t, lhs_extended, rhs_extended, reorder=True, splits=[4]
    )

    module = Module()
    module.add(ode_robertson)
    module.add(ode_baseline)
    module.add(ode_reordered)
    module.add(ode_reordered_with_splits)

    root_folder = tmpdir.join("root_folder").strpath
    lsoda_folder = tmpdir.join("lsoda").strpath

    wrapper = module.compile_and_load(
        root_folder=root_folder, lsoda_folder=lsoda_folder
    )

    tvec = 0.4 * 10 ** np.arange(0, 6)

    result_robertson, meta_robertson = wrapper.solve_fast_robertson(
        np.array(y0_vec), tvec, rtol=rtol, atol=np.array(atol)
    )

    result_reordered, meta_reordered = wrapper.solve_fast_reordered(
        np.array(y0_vec_extended),
        tvec,
        rtol=rtol,
        atol=np.array(atol_extended),
    )

    (
        result_reordered_with_splits,
        meta_reordered_with_splits,
    ) = wrapper.solve_fast_reordered_with_splits(
        np.array(y0_vec_extended),
        tvec,
        rtol=rtol,
        atol=np.array(atol_extended),
    )

    result_baseline, meta_baseline = wrapper.solve_fast_baseline(
        np.array(y0_vec_extended),
        tvec,
        rtol=rtol,
        atol=np.array(atol_extended),
    )

    assert np.allclose(result_baseline[:, :2:-1], result_robertson)
    assert np.allclose(result_baseline, result_reordered)
    assert np.allclose(result_baseline, result_reordered_with_splits)

    assert np.allclose(
        wrapper.reordered_dot(1.0, np.arange(6.0)),
        wrapper.baseline_dot(1.0, np.arange(6.0)),
    )



#!/usr/bin/env python

import glob
import os
import shutil
import subprocess

from setuptools._distutils import ccompiler

from .utils import create_folder_if_not_exists, get_platform

HERE = os.path.dirname(os.path.abspath(__file__))


def install_f2c_if_needed(cache_folder):
    folder = os.path.join(cache_folder, "sympy2c", get_platform(), "libf2c")
    lib_name = "libf2c.a"
    lib_path = os.path.join(folder, lib_name)
    if os.path.exists(lib_path):
        return lib_path

    create_folder_if_not_exists(folder)
    for path in glob.glob(os.path.join(HERE, "f2c_files", "*")):
        shutil.copy(path, folder)

    current_folder = os.getcwd()
    CC = os.environ.get("CC", ccompiler.new_compiler().executables["compiler"][0])

    try:
        os.chdir(folder)
        output = subprocess.check_output(
            [
                CC,
                "-O3",
                "-fPIC",
                "-c",
                "-DINTEGER_STAR_8=1",
                # "-w",
                "-Wfatal-errors",
                *glob.glob("*.c"),
            ],
            shell=False,
            universal_newlines=True,
        )
        print(output)

        output = subprocess.check_output(
            [
                "ar",
                "rcs",
                lib_name,
                *glob.glob("*.o"),
            ],
            shell=False,
            universal_newlines=True,
        )
        print(output)
    finally:
        os.chdir(current_folder)
    return lib_path



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

from sympy import BlockMatrix, Matrix, Number, Symbol, cse

from .symbol import symbols


def quick_inv(M):
    M = M.copy()
    n = M.shape[0]
    si = symbols(" ".join(f"_a{i}" for i in range(n * n)))
    cache = {}
    for i in range(n):
        for j in range(n):
            if not isinstance(M[i, j], (Symbol, Number)):
                symbol = si[i + n * j]
                cache[symbol] = M[i, j]
                M[i, j] = symbol

    return M.inv().subs(cache)


class RecursiveInverse:
    def __init__(self):
        self.xii = []
        self.next_symbol = self._symbol_generator()

    def _symbol_generator(self):
        i = 0
        while True:
            yield Symbol(f"__x{i}")
            i += 1

    def __call__(self, M):
        xii = []
        Minv = self._inv(xii, M)
        return xii, Minv

    def _inv(self, xii, M):
        n = M.shape[0]
        if n > 2:
            A = M[: n // 2, : n // 2]
            B = M[: n // 2, n // 2 :]
            C = M[n // 2 :, : n // 2]
            D = M[n // 2 :, n // 2 :]

            Ai = self._inv(xii, A)
            xi, (Ai,) = cse(Ai, self.next_symbol)
            xii.extend(xi)

            R = D - C * Ai * B

            Ri = self._inv(xii, R)
            xi, (Ri,) = cse(Ri, self.next_symbol)
            xii.extend(xi)

            M1 = Ai + Ai * B * Ri * C * Ai
            M2 = -Ai * B * Ri
            M3 = -Ri * C * Ai
            M4 = Ri

            Mi = Matrix(BlockMatrix([[M1, M2], [M3, M4]]))
        elif n == 1:
            Mi = M**-1
        else:
            Mi = quick_inv(M)
        return Mi


rec_inv = RecursiveInverse()



# This file is part of PyCosmo, a multipurpose cosmology calculation tool in Python.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.


import pickle
import warnings

import numpy as np
import pytest
import sympy as sp

from sympy2c import (
    ERROR,
    Alias,
    Checked,
    Function,
    Globals,
    IfThenElse,
    Integral,
    InterpolationFunction1D,
    Module,
    Symbol,
    Vector,
    isnan,
    symbols,
)


@pytest.fixture
def x():
    return Symbol("x")


@pytest.fixture
def y():
    return Symbol("y")


def test_finite_integral_qng_and_qag(tmpdir, compile_and_load):
    a, b, x = map(Symbol, "abx")
    integrand = 1 + (a + x) * x

    integral = Integral(integrand, x, 0, b)

    m = Module()
    m.add(Globals(b))
    m.add(Function("integral", integral, a))

    assert m.get_unique_id() is not None

    module = compile_and_load(m)
    assert module.get_unique_id() is not None
    # w = m.get_module_wrapper().setup_wrappers()
    # module = compile_and_load(m)
    module.set_quadpack_method("default", "QNG")
    module.set_eps_rel("default", 1e-5)
    module.set_eps_abs("default", 0)
    module.set_max_intervals("default", 1000)

    module.globals.b = 1
    assert abs(module.integral(0) - 1 - 1 / 3) < 1e-8
    assert module.get_abs_err("default") < 1e-13
    assert module.get_rel_err("default") < 1e-13

    assert abs(module.integral(1) - 1.5 - 1 / 3) < 1e-8
    assert module.get_abs_err("default") < 1e-13
    assert module.get_rel_err("default") < 1e-13

    module.set_quadpack_method("default", "QAG")
    module.globals.b = 1
    assert abs(module.integral(0) - 1 - 1 / 3) < 1e-8
    assert module.get_abs_err("default") < 1e-13
    assert module.get_rel_err("default") < 1e-13

    assert abs(module.integral(1) - 1.5 - 1 / 3) < 1e-8
    assert module.get_abs_err("default") < 1e-13
    assert module.get_rel_err("default") < 1e-13

    module.globals.b = None
    assert np.isnan(module.globals.b)


def test_infinite_integrals(tmpdir, compile_and_load):
    center = Symbol("center")
    x = Symbol("x")
    integrand = sp.exp(-sp.exp((x - center) ** 2))

    integral_1 = Integral(integrand, x, -sp.oo, sp.oo)
    integral_2 = Integral(integrand, x, center, sp.oo)
    integral_3 = Integral(integrand, x, -sp.oo, center)

    m = Module()
    m.add(Function("integral_1", integral_1, center))
    m.add(Function("integral_2", integral_2, center))
    m.add(Function("integral_3", integral_3, center))

    module = compile_and_load(m)
    module.set_quadpack_method("default", "QAGI")
    module.set_eps_rel("default", 1e-5)
    module.set_eps_abs("default", 0)
    module.set_max_intervals("default", 1000)

    expected_value = 0.5266003665440283

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        assert abs(module.integral_1(0) - expected_value) < 1e-10
        assert abs(module.integral_1(1) - expected_value) < 1e-10

        assert abs(module.integral_2(0) - expected_value / 2) < 1e-10
        assert abs(module.integral_2(1) - expected_value / 2) < 1e-10

        assert abs(module.integral_3(0) - expected_value / 2) < 1e-10
        assert abs(module.integral_3(1) - expected_value / 2) < 1e-10


def test_vector_symbol(tmpdir, compile_and_load):
    a = Vector("a", 3)
    x = Symbol("x")
    y = Symbol("y")

    expression = a[0] * Integral(a[1] * x + y, x, 0, a[2])

    m = Module()
    m.add(Function("fun", expression, a, y))
    assert m.get_unique_id() is not None

    module = compile_and_load(m)

    module.set_quadpack_method("default", "QAG")
    module.set_eps_rel("default", 1e-5)
    module.set_eps_abs("default", 0)
    module.set_max_intervals("default", 1000)

    assert module.fun(np.array([0.0, 0.0, 0.0]), 0.0) == 0
    assert module.fun(np.array([1.0, 1.0, 0.0]), 0.0) == 0
    assert abs(module.fun(np.array([1.0, 1.0, 1.0]), 0.0) - 0.5) < 1e-8
    assert abs(module.fun(np.array([2.0, 1.0, 1.0]), 0.0) - 1.0) < 1e-8


def test_vector_valued_functions(tmpdir, compile_and_load):
    a = Vector("a", 3)

    expressions = [a[0] * a[1], a[1] * a[2], a[0] * a[2]]
    f = Function("a", expressions, a)

    m = Module()
    m.add(f)
    module = compile_and_load(m)

    assert np.all(module.a(np.arange(1, 4, dtype=float)) == np.array([2.0, 6.0, 3.0]))

    from numba import njit

    aa = module.a

    @njit
    def compute():
        return aa(np.arange(1.0, 4.0))

    compute()


def test_alias(tmpdir, compile_and_load):
    x = Symbol("x")
    y = Symbol("y")
    z = Symbol("z")

    expressions = [x * y, y * z, x * z]

    a = Alias("a", x, y, z)

    f = Function("a", expressions, a)

    m = Module()
    m.add(f)
    module = compile_and_load(m)

    assert np.all(module.a(np.arange(1, 4, dtype=float)) == np.array([2.0, 6.0, 3.0]))


def test_interpolation(tmpdir, compile_and_load):
    x = Symbol("x")

    f = InterpolationFunction1D("f")

    m = Module()
    m.add(Function("f_interp", f(x) + x, x))
    assert m.get_unique_id() is not None

    module = compile_and_load(m)

    with pytest.raises(ArithmeticError):
        module.f_interp(1)
    assert module.get_last_error_message() == "you must call set_f_values first"

    x = np.linspace(0, 1, 100)
    y = x**2

    module.set_f_values(x, y + 1)
    assert module.f_interp(0.0) == 1

    assert module.f_interp(1.0) == 3

    assert np.all(module.ip_sorted_f(x) + x == module.f_interp(x))

    # check boundaries:
    assert np.allclose(module.ip_sorted_f(np.array([0.99])), [1.0 + 0.99**2], atol=1e-5)
    assert np.allclose(module.ip_sorted_f(np.array([1.0])), [1.0 + 1.0], atol=1e-5)
    assert np.allclose(module.ip_sorted_f(np.array([1e-6])), [1.0 + 1e-12], atol=1e-5)
    assert np.allclose(module.ip_sorted_f(np.array([0.0])), [1.0 + 0.0], atol=1e-5)

    assert np.all(np.isnan(module.ip_sorted_f(np.array([1.00000001]))))
    assert np.all(np.isnan(module.ip_sorted_f(np.array([-1e-10]))))

    # set aggain
    module.set_f_values(x, y)

    assert module.f_interp(0.0) == 0
    assert module.f_interp(1.0) == 2

    h = x[1] - x[0]
    assert abs(module.f_interp(0.5) - 0.75) <= h**2 / 4

    # continuation to the right is ok:
    module.f_interp(1.1)
    with pytest.raises(ArithmeticError):
        module.f_interp(-0.1)

    assert module.get_last_error_message() == "spline eval failed: input domain error"

    assert np.all(x == module.get_x_f())
    assert np.all(y == module.get_y_f())

    m_back = pickle.loads(pickle.dumps(module))
    assert np.all(x == m_back.get_x_f())
    assert np.all(y == m_back.get_y_f())

    assert m_back.f_interp(1.1) == module.f_interp(1.1)


def test_error_and_checked(tmpdir, compile_and_load):
    a = Symbol("a")
    expression = IfThenElse(a <= 0, ERROR("can not take log"), sp.log(a))

    f = Function("f", expression, a)

    a_checked = Checked(a >= 0, "a must not be negative")
    g = Function("g", sp.sqrt(a_checked), a)

    h = Function("h", IfThenElse(isnan(a), 1.0, 2.0), a)

    m = Module()
    m.add(f)
    m.add(g)
    m.add(h)
    module = compile_and_load(m)

    assert module.f(1.0) == 0.0
    assert module.g(0.0) == 0.0
    assert module.g(1.0) == 1.0
    assert module.g(4.0) == 2.0

    assert module.h(np.nan) == 1.0
    assert module.h(1.0) == 2.0

    with pytest.raises(ArithmeticError):
        module.f(0)

    assert "can not take log" in module.get_last_error_message()

    with pytest.raises(ArithmeticError):
        module.g(-1)

    assert "a must not be negative" in module.get_last_error_message()


def test_integrals(compile_and_load):
    a, b, c, d, e = symbols("a b c d e")

    m = Module()
    ii = Integral(a + b * c, c, 1, a)
    f = Function("f", ii, a, b)
    m.add(f)

    jj = Integral(a + b * c, c, d, e)
    g = Function("g", jj, a, b, d, e)
    m.add(g)
    compile_and_load(m)



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from functools import partial

from .compiler import LoadedModule, build_folder, compile_if_needed_and_load
from .create_curry_class import create_interface_onary
from .utils import Hasher, align, concat_generator_results
from .wrapper import WrapperBase


def _print_code_fragment(code, line_no, e, stream=sys.stderr, extra_lines=5):
    for i, line in enumerate(code.split("\n"), 1):
        if i < line_no - extra_lines or i > line_no + extra_lines:
            continue
        if i == line_no:
            print(">>>", str(e) + ":", file=stream)
        print("{:3d}".format(i), line.rstrip(), file=stream)


class Module(object):
    def __init__(self):
        from .globals import Globals

        self.globals_ = Globals()
        self.functions = []
        self.python_functions = []
        self.odes = []
        self.fast_odes = []
        self.combined_odes = []
        self._reset_unique_id()

    def _reset_unique_id(self):
        self._unique_id = None

    def add(self, what):
        from .function import Function
        from .globals import Globals
        from .ode import Ode
        from .ode_combined import OdeCombined
        from .ode_fast import OdeFast
        from .python_function import PythonFunction

        self._reset_unique_id()
        if isinstance(what, Function):
            self._add_function(what)
        elif isinstance(what, Globals):
            self._add_globals(what)
        elif isinstance(what, Ode):
            self._add_ode(what)
        elif isinstance(what, OdeFast):
            self._add_ode_fast(what)
        elif isinstance(what, OdeCombined):
            self._add_ode_combined(what)
        elif isinstance(what, PythonFunction):
            self._add_python_function(what)
        else:
            raise ValueError("don't know how to add {!r}".format(what))

    def get_unique_id(self):
        if self._unique_id is None:
            md5 = Hasher()
            for di in self.functions:
                md5.update(di._name, di.get_unique_id())
            for di in self.odes:
                md5.update(di.name, di.get_unique_id())
            for di in self.fast_odes:
                md5.update(di.name, di.get_unique_id())
            for di in self.combined_odes:
                md5.update(di.name, di.get_unique_id())
            for di in self.python_functions:
                md5.update(di.name, di.get_unique_id())

            from . import __version__

            md5.update("globals", self.globals_.get_unique_id())
            version = "_".join(map(str, __version__))
            self._unique_id = version + "_" + md5.hexdigest()
        return self._unique_id

    def _add_function(self, f):
        common = set(self.globals_) & set(f.arguments)
        assert not common, "symbols(s) {} already declared as globals".format(
            ", ".join(map(str, common))
        )
        self.functions.append(f)

    def _add_ode(self, ode):
        self.odes.append(ode)

    def _add_ode_fast(self, ode):
        self.fast_odes.append(ode)

    def _add_ode_combined(self, ode):
        self.combined_odes.append(ode)

    def _add_python_function(self, f):
        self.python_functions.append(f)

    def _add_globals(self, globals_):
        for symbol in globals_.variables:
            self.globals_.add_variable(symbol)

    @staticmethod
    def _setup_ns():
        m = Module()

        def add_function(name, expression, *args):
            f = Function(name, expression, *args)
            m.add(f)
            return f

        def add_ode(name, tvar, lhs, rhs, **kw):
            ode = Ode(name, tvar, lhs, rhs, **kw)
            m.add(ode)
            return ode

        def add_ode_fast(name, tvar, lhs, rhs, **kw):
            ode = OdeFast(name, tvar, lhs, rhs, **kw)
            m.add(ode)
            return ode

        def add_ode_combined(*a, **kw):
            ode = OdeCombined(*a, **kw)
            m.add(ode)
            return ode

        def decl(*names, ns=None):
            for name in names:
                exec("{name} = Symbol('{name}')".format(name=name), ns)

        from .expressions import Max, Min, isnan
        from .function import Alias, Function
        from .integral import ERROR, Checked, IfThenElse, Integral
        from .interpolation import InterpolationFunction1D
        from .ode import Ode
        from .ode_combined import OdeCombined
        from .ode_fast import OdeFast
        from .symbol import Symbol
        from .vector import Vector

        ns = {
            "Function": add_function,
            "Integral": Integral,
            "InterpolationFunction1D": InterpolationFunction1D,
            "Symbol": Symbol,
            "Alias": Alias,
            "Min": Min,
            "Max": Max,
            "isnan": isnan,
            "Vector": Vector,
            "IfThenElse": IfThenElse,
            "ERROR": ERROR,
            "Checked": Checked,
            "Ode": add_ode,
            "OdeFast": add_ode_fast,
            "OdeCombined": add_ode_combined,
            "globals_": [],
        }
        return ns, decl, m

    @staticmethod
    def load_sympy_code(path, locals_=None):
        if locals_ is None:
            locals_ = {}

        ns, decl, _ = Module._setup_ns()
        __builtins__.update(ns)
        __builtins__["decl"] = partial(decl, ns=__builtins__)
        __builtins__.update(locals_)
        import importlib
        import os

        try:
            sys.path.insert(0, os.path.dirname(path))
            importlib.__import__(os.path.basename(path)[:-3], globals(), ns)
        finally:
            del sys.path[0]

    @staticmethod
    def parse_sympy_code(code, locals_=None):
        from .globals import Globals

        if locals_ is None:
            locals_ = {}

        ns, decl, m = Module._setup_ns()
        ns["decl"] = partial(decl, ns=ns)

        ns.update(locals_)

        try:
            exec(code, ns)
        except Exception as e:
            tb = sys.exc_info()[-1]
            tb_next = tb.tb_next
            if tb_next is not None:
                line_no = tb_next.tb_lineno
                _print_code_fragment(code, line_no, e)
            raise e
        globals_ = ns["globals"]
        m.add(Globals(*globals_))

        return m, ns

    @staticmethod
    def parse_sympy_file(path, locals_=None):
        with open(path, "r") as fh:
            return Module.parse_sympy_code(fh.read(), locals_)

    def get_module_wrapper(self):
        return ModuleWrapper(self)

    def compile_and_load(
        self,
        root_folder=None,
        lsoda_folder=None,
        gsl_root=None,
        compilation_flags=None,
    ):
        wrapper = ModuleWrapper(self)
        wrapper.setup_wrappers(True, None)
        return wrapper.compile_and_load(
            root_folder, lsoda_folder, gsl_root, compilation_flags
        )

    def recompile_and_load(
        self,
        root_folder=None,
        lsoda_folder=None,
        gsl_root=None,
        compilation_flags=None,
        force=False,
    ):
        from .ode_fast import read_known_traces, read_new_traces, write_known_traces

        found_new_traces = False
        for ode in self.fast_odes:
            N = len(ode.rhs)
            splits = ode.splits
            uid = ode.get_unique_id()
            latest_traces = read_known_traces(uid)
            new_traces = read_new_traces(uid)
            if not new_traces:
                continue
            if not latest_traces:
                sizes = (
                    s
                    for s in [splits[0]]
                    + [s1 - s0 for (s0, s1) in zip(splits, splits[1:])]
                    + [N - splits[-1]]
                    if s
                )
                latest_traces = {i: [list(range(s))] for i, s in enumerate(sizes)}
            for level, traces in new_traces.items():
                for trace in traces:
                    if trace not in latest_traces[level]:
                        latest_traces[level].append(trace)
                        found_new_traces = True
            write_known_traces(uid, latest_traces)
        if force or found_new_traces:
            return self.compile_and_load(
                root_folder, lsoda_folder, gsl_root, compilation_flags
            )
        else:
            wrapper = ModuleWrapper(self)
            wrapper.setup_wrappers(True, None)
            return LoadedModule(build_folder(wrapper, compilation_flags, root_folder))


class ModuleWrapper(WrapperBase):
    def __init__(self, module):
        from . import __version__

        self.__version__ = (
            __version__
        )  # for correct caching of generated and compiled code
        self.functions = module.functions
        self.globals_ = module.globals_
        self.odes = module.odes[:]
        self.fast_odes = module.fast_odes[:]
        self.combined_odes = module.combined_odes[:]
        self.python_functions = module.python_functions[:]

        from .visitor import Sp2CVisitor

        self.visitor = Sp2CVisitor(self.globals_)

        self.function_wrappers = None
        self.ode_wrappers = None
        self.fast_ode_wrappers = None
        self._unique_id_0 = None
        self._unique_id = None

    def setup_wrappers(self, compile_fast_odes=False, traces=None):
        from .function import function_wrapper_factory

        self.function_wrappers = [
            function_wrapper_factory(function, self.globals_, self.visitor)
            for function in self.functions
        ]
        self.ode_wrappers = [
            ode.wrapper(ode, self.globals_, self.visitor) for ode in self.odes
        ]

        self.combined_ode_wrappers = [
            ode.wrapper(ode, self.globals_, self.visitor) for ode in self.combined_odes
        ]

        self.python_function_wrappers = [f.wrapper() for f in self.python_functions]

        if traces is None:
            traces = {}

        if compile_fast_odes:
            self.fast_ode_wrappers = [
                ode.wrapper(ode, self.globals_, self.visitor, traces.get(ode.name))
                for ode in self.fast_odes
            ]
        else:
            self.fast_ode_wrappers = []

        for w in self.function_wrappers + self.ode_wrappers + self.fast_ode_wrappers:
            w.determine_required_extra_wrappers()

        return self

    def setup_code_generation(self):
        for wrapper in (
            self.function_wrappers
            + self.ode_wrappers
            + self.fast_ode_wrappers
            + self.combined_ode_wrappers
            + self.python_function_wrappers
            + list(self.visitor.extra_wrappers.values())
        ):
            wrapper.setup_code_generation()

    def get_unique_id(self):
        if self.function_wrappers is None:
            raise RuntimeError("you must first call setup_wrappers()")

        if self._unique_id is not None:
            return self._unique_id

        if self._unique_id_0 is None:
            md5 = Hasher()
            for di in (
                self.function_wrappers
                + self.ode_wrappers
                + self.python_function_wrappers
            ):
                uid = di.get_unique_id()
                md5.update(di.name, uid)
            self._unique_id_0 = md5.hexdigest()

        md5 = Hasher()
        md5.update("unique_id_0", self._unique_id_0)

        for di in self.fast_ode_wrappers + self.combined_ode_wrappers:
            uid = di.get_unique_id()
            md5.update(di.name, uid)

        md5.update("globals", self.globals_.get_unique_id())
        self._unique_id = md5.hexdigest()
        return self._unique_id

    @concat_generator_results
    def c_header(self):
        yield _gsl_related_header()
        yield _macros()
        yield _lsoda_header()
        yield _ode_fast_general_header()
        yield self.globals_.c_header()

        for wrapper in (
            self.function_wrappers
            + self.ode_wrappers
            + self.fast_ode_wrappers
            + self.combined_ode_wrappers
            + self.python_function_wrappers
            + list(self.visitor.extra_wrappers.values())
        ):
            yield wrapper.c_header()

    @concat_generator_results
    def c_code(self, header_file_path):
        yield """#include "{header_file_path}"\n""".format(
            header_file_path=header_file_path
        )
        yield self.globals_.c_code(header_file_path)

        one_ary_interface_class_code = create_interface_onary()

        yield one_ary_interface_class_code

        yield from _gsl_related_c_code()
        yield from _lsoda_related_c_code()
        yield from _lsoda_fast_related_c_code()

        for wrapper in (
            self.function_wrappers
            + self.ode_wrappers
            + self.fast_ode_wrappers
            + self.combined_ode_wrappers
            + self.python_function_wrappers
            + list(self.visitor.extra_wrappers.values())
        ):
            yield wrapper.c_code(header_file_path)

    @concat_generator_results
    def cython_code(self, header_file_path):
        yield from _general_cython_code(self._unique_id)
        yield from _gsl_related_cython_code(header_file_path)
        yield from _lsoda_fast_related_cython_code()

        yield self.globals_.cython_code(header_file_path)

        for wrapper in (
            self.function_wrappers
            + self.ode_wrappers
            + self.fast_ode_wrappers
            + self.combined_ode_wrappers
            + self.python_function_wrappers
            + list(self.visitor.extra_wrappers.values())
        ):
            yield wrapper.cython_code(header_file_path)

        yield from self.fast_ode_unique_id_function()

        yield self._pickling_code()

    def _pickling_code(self):
        names = []
        from .interpolation import InterpolationFunction1DWrapper

        for name, w in self.visitor.extra_wrappers.items():
            if isinstance(w, InterpolationFunction1DWrapper):
                names.append(str(name))

        return align(
            f"""
                |def __getstate__():
                |    data = [(name,
                |             eval("get_x_" + name + "()"),
                |             eval("get_y_" + name + "()"))
                |    for name in {names!r}]
                |    # print("get state", data)
                |    return data
                |
                |def __setstate__(data):
                |    # print("set state", data)
                |    for name, x, y in data:
                |         if x is not None:
                |             eval("set_" + name + "_values(x, y)")
                """
        )

    def fast_ode_unique_id_function(self):
        unique_ids = {
            wrapper.name: wrapper.ode.get_unique_id()
            for wrapper in self.fast_ode_wrappers
        }

        yield align(
            f"""
                |def get_fast_ode_unique_ids():
                |    return {unique_ids!r}
                """
        )
        yield align(
            """
            |def update_new_traces(unique_id, new_traces):
            |    from sympy2c.ode_fast import update_new_traces
            |    update_new_traces(unique_id, new_traces)
        """
        )

    def determine_required_extra_wrappers(self):
        pass

    def compile_and_load(
        self, root_folder=None, lsoda_folder=None, gsl_root=None, compilation_flags=None
    ):
        return compile_if_needed_and_load(
            self,
            root_folder,
            lsoda_folder,
            gsl_root,
            compilation_flags,
        )


def _macros():
    return align(
        """
    |inline double _min(double a, double b) {
    |       return (a < b) ? a : b;
    |}
    |
    |inline double _max(double a, double b) {
    |       return (a > b) ? a : b;
    |}
            """
    )


def _lsoda_header():
    return align(
        """
    |extern "C" {
    |    typedef void (*_lsoda_f) (double t, double *y, double *ydot, void *data);
    |    typedef void (*_lsoda_jac) (double t, double *y, double *jac, void *data);
    |
    |    void lsoda(_lsoda_f f, _lsoda_jac jac, int neq, double *y, double *t,
    |               double tout, int itol, double *rtol, double *atol, int itask,
    |               int *istate, int iopt, int jt, int iwork1, int iwork2, int iwork5,
    |               int iwork6, int iwork7, int iwork8, int iwork9, double rwork1,
    |               double rwork5, double rwork6, double rwork7, void *data);
    |
    |    void n_lsoda_terminate();
    |}
    """
    )


def _ode_fast_general_header():
    return align(
        """
    |#include <set>
    |#include <map>
    |#include <vector>
    |typedef std::set<std::vector<int>> TRACES;
    |typedef std::map<string, std::map<int, TRACES>> PT;
    |
    |PT get_new_traces();
    """
    )


def _general_cython_code(unique_id):
    from . import __version__

    yield align(
        """
           |# distutils: language = c++
           |
           |cimport libc.stdio as stdio
           |from libc.stdio cimport printf, fflush, stdout
           |from libc.stdlib cimport malloc
           |from posix.time cimport timeval, gettimeofday
           |from libcpp.vector cimport vector
           |from libcpp.set cimport set as cset
           |from libcpp.string cimport string
           |from libcpp.map cimport map
           |from libcpp.unordered_map cimport unordered_map
           |from cython.operator cimport dereference as deref, preincrement as inc
           |
           |cimport cython
           |
           |sympy2c_version = "{version}"
           |
           |cimport numpy as np
           |import numpy as np
           |import ctypes
           |
           |_api = {{}}
           |
           |cdef long microtime():
           |     cdef timeval time
           |     gettimeofday(&time, NULL)
           |     return (<unsigned long long>time.tv_sec * 1000000) + time.tv_usec
           |
           |import warnings
           |
           |def get_unique_id():
           |    return "{unique_id}"
           |
           """.format(unique_id=unique_id, version=__version__)
    )


def _gsl_related_header():
    return align(
        """
    |#include <cmath>
    |#include <cstdio>
    |#include <cstring>
    |#include <unordered_map>
    |#include <string>
    |
    |#include <gsl/gsl_integration.h>
    |#include <gsl/gsl_errno.h>
    |#include <gsl/gsl_spline.h>
    |#include <gsl/gsl_sf.h>
    |
    |
    |using namespace std;
    |
    |enum QuadPackMethod {
    |    QNG,
    |    QAG,
    |    QAGI
    |};
    |
    |extern "C" const char * get_last_error_message();
    |extern "C" const char * get_last_warning_message();
    |extern "C" void clear_last_error_message();
    |extern "C" void clear_last_warning_message();
    |extern "C" void set_eps_rel(string id, double value);
    |extern "C" void set_eps_abs(string id, double value);
    |extern "C" double get_rel_err(string id);
    |extern "C" double get_abs_err(string id);
    |extern "C" void set_eps_abs(string id, double value);
    |extern "C" void set_max_intervals(string id, size_t value);
    |extern "C" void set_quadpack_method(string id, QuadPackMethod m);
    |extern "C" void set_gsl_error_handler();
    |
    """
    )


def _gsl_related_c_code():
    yield align(
        """
        |#define BUFFER_SIZE_ERROR_MESSAGE 9999
        |
        |#define APPEND(buffer, buffer_size, what) do {\\
        |    strncat(buffer, what, buffer_size - strlen(buffer) - 1);\\
        |    } while(0)
        |
        |static char last_error_message[BUFFER_SIZE_ERROR_MESSAGE] = "";
        |static char last_warning_message[BUFFER_SIZE_ERROR_MESSAGE] = "";
        |
        |void set_error_message_gsl(const char * prefix, int code)
        |{
        |    clear_last_error_message();
        |    APPEND(last_error_message, BUFFER_SIZE_ERROR_MESSAGE, prefix);
        |    APPEND(last_error_message, BUFFER_SIZE_ERROR_MESSAGE, ": ");
        |    APPEND(last_error_message, BUFFER_SIZE_ERROR_MESSAGE, gsl_strerror(code));
        |}
        |
        |double set_error_message(const char * message)
        |{
        |    clear_last_error_message();
        |    APPEND(last_error_message, BUFFER_SIZE_ERROR_MESSAGE, message);
        |    return NAN;
        |}
        |
        |void set_warning_message(const char * message)
        |{
        |    clear_last_error_message();
        |    APPEND(last_warning_message, BUFFER_SIZE_ERROR_MESSAGE, message);
        |}
        |
        |static unordered_map<string, double> id_to_eps_rel = {{"default", 1e-6}};
        |static unordered_map<string, double> id_to_eps_abs = {{"default", 0.0 }};
        |static unordered_map<string, size_t> id_to_max_intervals = {{"default", 1000}};
        |static unordered_map<string, QuadPackMethod> id_to_quadpack_method
        |                                             = {{"default", QAG}};
        |
        |static unordered_map<string, double> id_to_rel_err = {{ }};
        |static unordered_map<string, double> id_to_abs_err = {{ }};

        |extern "C" {
        |    const char * get_last_error_message() { return last_error_message; }
        |    const char * get_last_warning_message() { return last_warning_message; }
        |    void clear_last_error_message() { last_error_message[0] = 0; }
        |    void clear_last_warning_message() { last_warning_message[0] = 0; }
        |
        |    void set_eps_rel(string id, double value) { id_to_eps_rel[id] = value; }
        |    void set_eps_abs(string id, double value) { id_to_eps_abs[id] = value; }
        |
        |    void set_rel_err(string id, double value) { id_to_rel_err[id] = value; }
        |    void set_abs_err(string id, double value) { id_to_abs_err[id] = value; }
        |
        |    double get_rel_err(string id) {
        |        return id_to_rel_err.count(id) > 0 ? id_to_rel_err[id]: -1.0;
        |    }
        |    double get_abs_err(string id) {
        |        return id_to_abs_err.count(id) > 0 ? id_to_abs_err[id]: -1.0;
        |    }
        |
        |    void set_max_intervals(string id, size_t value) {
        |        id_to_max_intervals[id] = value;
        |    }
        |    void set_quadpack_method(string id, QuadPackMethod m) {
        |         id_to_quadpack_method[id] = m;
        |    }
        |}
        |
        |void handler (const char * reason, const char * file, int line, int gsl_errno)
        |{
        |
        |    last_error_message[0] = 0;
        |    char buffer[200];
        |    snprintf(buffer, 200, " (line %d): %s, ", line, reason);
        |    APPEND(last_error_message, BUFFER_SIZE_ERROR_MESSAGE, file);
        |    APPEND(last_error_message, BUFFER_SIZE_ERROR_MESSAGE, buffer);
        |    APPEND(last_error_message, BUFFER_SIZE_ERROR_MESSAGE,
        |           gsl_strerror(gsl_errno));
        |
        |}
        |
        |void set_gsl_error_handler() {
        |   gsl_set_error_handler(handler);
        |}
        |
        |inline double gsl_function_eval(double x, void * curried_function)
        |{
        |   OneAryFunction * func = (OneAryFunction *) curried_function;
        |   return (*func)(x);
        |}
        |
        |#define BUFFER_SIZE_SMALL 100
        |
        |#define REPORT_ERROR(fmt, arg, code) do { \\
        |    snprintf(buffer, BUFFER_SIZE_SMALL, fmt, arg); \\
        |    set_error_message_gsl(buffer, code); \\
        |   } while(0)
        |
        |#define REPORT_WARNING(fmt, arg) do { \\
        |    snprintf(buffer, BUFFER_SIZE_SMALL, fmt, arg); \\
        |    set_warning_message(buffer); \\
        |   } while(0)
        |
        |double quadpack_integration(OneAryFunction *f, const char *id,
        |                            double low, double high) {
        |    double result, error, eps_rel, eps_abs;
        |    size_t max_intervals;
        |    QuadPackMethod method;
        |    int err_code;
        |    unordered_map<string, double>::const_iterator found_eps;
        |    unordered_map<string, size_t>::const_iterator found_max_intervals;
        |    unordered_map<string, QuadPackMethod>::const_iterator found_method;
        |    char buffer[BUFFER_SIZE_SMALL];
        |
        |    found_method = id_to_quadpack_method.find(id);
        |    if (found_method != id_to_quadpack_method.end())
        |         method = found_method->second;
        |    else {
        |         REPORT_ERROR("no quadpack method set for id '%s'", id, GSL_EINVAL);
        |         return GSL_EINVAL;
        |    }
        |
        |    if (std::isinf(low) || std::isinf(high)) {
        |        if (method != QAGI) {
        |            //REPORT_WARNING(
        |            //    "switched to QAGI for infinite limit(s) for id '%s'", id
        |            //);
        |            method = QAGI;
        |        }
        |    }
        |    else if (method == QAGI) {
        |       REPORT_ERROR("you must not use QAGI for finite limit(s) for id '%s'",
        |                     id, GSL_EINVAL);
        |       return GSL_EINVAL;
        |
        |    }
        |
        |    found_eps = id_to_eps_rel.find(id);
        |    if (found_eps != id_to_eps_rel.end())
        |         eps_rel = found_eps->second;
        |    else {
        |         REPORT_ERROR("no eps_rel value set for id '%s'", id, GSL_EINVAL);
        |         return GSL_EINVAL;
        |    }
        |
        |    found_eps = id_to_eps_abs.find(id);
        |    if (found_eps != id_to_eps_abs.end())
        |          eps_abs = found_eps->second;
        |    else {
        |         REPORT_ERROR("no eps_abs value set for id '%s'", id, GSL_EINVAL);
        |         return GSL_EINVAL;
        |    }
        |
        |    gsl_function F;
        |    F.function = gsl_function_eval;
        |    F.params = (void *)f;
        |
        |    gsl_integration_workspace *w;
        |    /* gsl_set_error_handler_off(); */
        |    clear_last_error_message();
        |
        |    if (method == QAG) {
        |        found_max_intervals = id_to_max_intervals.find(id);
        |        if (found_max_intervals != id_to_max_intervals.end())
        |             max_intervals = found_max_intervals->second;
        |        else {
        |            REPORT_ERROR(
        |                "no max_intervals value set for id '%s'", id, GSL_EINVAL
        |            );
        |            return GSL_EINVAL;
        |        }
        |
        |        w = gsl_integration_workspace_alloc(max_intervals);
        |
        |        err_code = gsl_integration_qag(&F, low, high, eps_abs, eps_rel,
        |                                       max_intervals, GSL_INTEG_GAUSS31, w,
        |                                       &result, &error);
        |
        |        if (err_code) {
        |            set_error_message_gsl("qag from gsl returned", err_code);
        |            set_abs_err(id, -1.0);
        |            set_rel_err(id, -1.0);
        |            }
        |        else {
        |            set_abs_err(id, error);
        |            set_rel_err(id, error / result);
        |        }
        |        gsl_integration_workspace_free(w);
        |
        |    } else if (method == QAGI) {
        |
        |        double sign = 1.0;
        |        double temp;
        |        if (low > high) {
        |            sign = -1.0;
        |            temp = low;
        |            low = high;
        |            high= temp;
        |        }
        |
        |        found_max_intervals = id_to_max_intervals.find(id);
        |        if (found_max_intervals != id_to_max_intervals.end())
        |             max_intervals = found_max_intervals->second;
        |        else {
        |            REPORT_ERROR(
        |                "no max_intervals value set for id '%s'", id, GSL_EINVAL
        |            );
        |            return GSL_EINVAL;
        |        }
        |
        |        w = gsl_integration_workspace_alloc(max_intervals);
        |
        |        if (std::isinf(low) && std::isinf(high)) {
        |            err_code = gsl_integration_qagi(&F, eps_abs, eps_rel,
        |                                            max_intervals,w, &result, &error);
        |            if (err_code)
        |                set_error_message_gsl("qagi from gsl returned", err_code);
        |        }
        |        else if (std::isinf(low)) {
        |            err_code = gsl_integration_qagil(&F, high, eps_abs, eps_rel,
        |                                           max_intervals, w, &result, &error);
        |            if (err_code)
        |                set_error_message_gsl("qagil from gsl returned", err_code);
        |        }
        |        else {
        |            err_code = gsl_integration_qagiu(&F, low, eps_abs, eps_rel,
        |                                           max_intervals, w, &result, &error);
        |            if (err_code)
        |                set_error_message_gsl("qagiu from gsl returned", err_code);
        |        }
        |        result *= sign;
        |
        |        if (err_code) {
        |            set_error_message_gsl("qagi from gsl returned", err_code);
        |            set_abs_err(id, -1.0);
        |            set_rel_err(id, -1.0);
        |            }
        |        else {
        |            set_abs_err(id, error);
        |            set_rel_err(id, error / result);
        |        }
        |
        |        gsl_integration_workspace_free(w);
        |
        |    } else {
        |
        |        size_t neval;
        |        err_code = gsl_integration_qng(&F, low, high, eps_abs, eps_rel,
        |                                       &result, &error, &neval);
        |
        |        if (err_code) {
        |            set_error_message_gsl("qagi from gsl returned", err_code);
        |            set_abs_err(id, -1.0);
        |            set_rel_err(id, -1.0);
        |            }
        |        else {
        |            set_abs_err(id, error);
        |            set_rel_err(id, error / result);
        |        }
        |    }
        |
        |    return result;
        |}
        |
        |typedef struct
        |   {
        |   double * c;
        |   double * g;
        |   double * diag;
        |   double * offdiag;
        |} cspline_state_t;
        |
        |static inline void
        |coeff_calc (const double c_array[], double dy, double dx, size_t index,
        |            double * b, double * c, double * d)
        |{
        |    const double c_i = c_array[index];
        |    const double c_ip1 = c_array[index + 1];
        |    *b = (dy / dx) - dx * (c_ip1 + 2.0 * c_i) / 3.0;
        |    *c = c_i;
        |    *d = (c_ip1 - c_i) / (3.0 * dx);
        |}
        |
        |static
        |void
        |cspline_eval_vector_sorted (const void * vstate,
        |            const double x_array[], const double y_array[], size_t size,
        |            double *x, int n,
        |            double *y)
        |{
        |    const cspline_state_t *state = (const cspline_state_t *) vstate;
        |
        |    double x_lo, x_hi;
        |    double dx;
        |    double xi;
        |    size_t index = 0;
        |
        |    index = gsl_interp_bsearch (x_array, x[0], 0, size - 1);
        |
        |    for (size_t ii=0; ii < n; ++ii) {
        |        xi = x[ii];
        |        y[ii] = NAN;
        |
        |        if (xi < x_array[0]) continue;
        |        if (xi > x_array[size - 1]) continue;
        |        while ((xi < x_array[index] || xi >= x_array[index + 1])
        |               && index < size - 2)
        |                   index ++;
        |       if (index < size - 2)
        |           if (xi < x_array[index] || xi >= x_array[index + 1]) {
        |              // printf("no match %d %lf %lf %lf\\n", index, xi, x_array[index], x_array[index + 1]);
        |              continue;
        |        }
        |
        |        if (index == size - 2)
        |           if (xi < x_array[index] || xi > x_array[index + 1]) {
        |              // printf("no match %d %lf %lf %lf\\n", index, xi, x_array[index], x_array[index + 1]);
        |              continue;
        |        }
        |
        |        /* evaluate */
        |        x_hi = x_array[index + 1];
        |        x_lo = x_array[index];
        |        dx = x_hi - x_lo;
        |        if (dx > 0.0) {
        |            const double y_lo = y_array[index];
        |            const double y_hi = y_array[index + 1];
        |            const double dy = y_hi - y_lo;
        |            double delx = xi - x_lo;
        |            double b_i, c_i, d_i;
        |            coeff_calc(state->c, dy, dx, index,  &b_i, &c_i, &d_i);
        |            // printf("%d %lf %lf %lf %lf %lf\\n", index, dy, dx, b_i, c_i, d_i);
        |            y[ii] = y_lo + delx * (b_i + delx * (c_i + delx * d_i));
        |        }
        |    }
        |}
        """
    )


def _gsl_related_cython_code(header_file_path):
    yield align(
        """
           |cdef extern from "{header_file_path}":
           |
           |    enum QuadPackMethod:
           |        QNG
           |        QAG
           |        QAGI
           |
           |cdef extern from "{header_file_path}":
           |
           |    char * c_get_last_error_message "get_last_error_message"()
           |    char * c_get_last_warning_message "get_last_warning_message"()
           |    void clear_last_error_message()
           |    void clear_last_warning_message()
           |    void set_gsl_error_handler()
           |    void c_set_eps_rel "set_eps_rel"(string, double)
           |    void c_set_eps_abs "set_eps_abs"(string, double)
           |
           |    void c_set_rel_err "set_rel_err"(string, double)
           |    void c_set_abs_err "set_abs_err"(string, double)
           |    double c_get_rel_err "get_rel_err"(string)
           |    double c_get_abs_err "get_abs_err"(string)
           |
           |    void c_set_max_intervals "set_max_intervals"(string, size_t)
           |    void c_set_quadpack_method "set_quadpack_method"(string, QuadPackMethod)
           |
           |def get_last_error_message():
           |    return str(c_get_last_error_message(), "utf-8")
           |
           |def get_last_warning_message():
           |    return str(c_get_last_warning_message(), "utf-8")
           |
        """.format(header_file_path=header_file_path)
    )

    yield align(
        """
        |def set_eps_rel(str id, double value):
        |    c_set_eps_rel(id.encode("ascii"), value)
        |
        |def set_eps_abs(str id, double value):
        |    c_set_eps_abs(id.encode("ascii"), value)
        |
        |def set_max_intervals(str id, size_t value):
        |    c_set_max_intervals(id.encode("ascii"), value)
        |
        |def set_quadpack_method(str id, str value):
        |    if value == "QNG":
        |       c_set_quadpack_method(id.encode("ascii"), QuadPackMethod.QNG)
        |    elif value == "QAG":
        |       c_set_quadpack_method(id.encode("ascii"), QuadPackMethod.QAG)
        |    elif value == "QAGI":
        |       c_set_quadpack_method(id.encode("ascii"), QuadPackMethod.QAGI)
        |    else:
        |       raise ValueError("unknown quadpack method {}".format(value))
        |
        |set_gsl_error_handler()
        |
        |def get_abs_err(str id):
        |    return c_get_abs_err(id.encode("ascii"))
        |def get_rel_err(str id):
        |    return c_get_rel_err(id.encode("ascii"))
        """
    )


def _lsoda_fast_related_c_code():
    yield align(
        r"""
            |double SEC_FACTOR=1;
            |extern "C" void _set_sec_factor(double sf) { SEC_FACTOR = sf; }
            |
            |PT _new_traces;
            |PT get_new_traces() { return _new_traces; }
            |
            |#include <chrono>
            |typedef std::chrono::high_resolution_clock clock_;
            |static std::chrono::time_point<clock_> m_beg;
            |
            |void reset_timer() { m_beg = clock_::now(); }
            |
            |float time_delta() {
            |    return 1e-9 * std::chrono::duration_cast<std::chrono::nanoseconds>(
            |                clock_::now() - m_beg).count();
            |}
            """
    )


def _lsoda_related_c_code():
    yield align(
        r"""
        |
        |#include<cstdio>
        |#include<cmath>
        |#include<list>
        |#include<vector>
        |#include<array>
        |
        |#define doublereal double
        |#define integer long int
        |
        |extern "C" struct {
        |    doublereal rowns[209], ccmax, el0, h, hmin, hmxi, hu, rc, tn, uround;
        |    integer illin, init, lyh, lewt, lacor, lsavf, lwm, liwm, mxstep, mxhnil,
        |    nhnil, ntrep, nslast, nyh, iowns[6], icf, ierpj, iersl, jcur,
        |    jstart, kflag, l, meth, miter, maxord, maxcor, msbp, mxncf, n, nq,
        |    nst, nfe, nje, nqu;
        |} ls0001_;
        |struct _solver_change {
        |    _solver_change(double _t,
        |                  int _from,
        |                  int _to): t(_t), from(_from), to(_to) {};
        |    double t;
        |    int from, to;
        |};

    """
    )


def _lsoda_fast_related_cython_code():
    yield align(
        """
        |ctypedef int (*S_fp)(...)
        |ctypedef int (*U_fp)(...)
        |ctypedef void (*V_fp)(...)
        |ctypedef double (*D_fp)(...)
        |
        |cdef extern:
        |    int lsodafast_(S_fp f, U_fp econ, long int *neq, double *y,
        |        double *t, double *tout, long int *itol, double *rtol,
        |        double *atol, long int *itask, long int *istate, long int *iopt,
        |        double *rwork, long int *lrw, long int *iwork, long int *liw,
        |        U_fp jac, long int *jt, V_fp lu_solver, D_fp fnorm_computation,
        |        V_fp step_size_callback, long *handle) nogil
        |
        |    cdef struct ls0001_struct:
        |        double rowns[209], ccmax, el0, h, hmin, hmxi, hu, rc, tn, uround
        |        long illin, init, lyh, lewt, lacor, lsavf, lwm, liwm, mxstep, mxhnil
        |        long nhnil, ntrep, nslast, nyh, iowns[6], icf, ierpj, iersl, jcur
        |        long jstart, kflag, l, meth, miter, maxord, maxcor, msbp, mxncf, n, nq
        |        long nst, nfe, nje, nqu
        |
        |    ls0001_struct ls0001_
        |
        |    void _set_sec_factor(double)
        |
        |cdef extern from "functions.hpp":
        |  map[string, map[int, cset[vector[int]]]] c_get_new_traces "get_new_traces" ()
        |
        |cpdef get_new_traces(str id_):
        |   cdef string id__ = id_.encode("utf-8")
        |
        |   cdef dict result = {}
        |
        |   cdef map[int, cset[vector[int]]] traces = c_get_new_traces()[id__];
        |   cdef map[int, cset[vector[int]]].iterator it_levels
        |   cdef cset[vector[int]].iterator it_level_traces
        |
        |   result = {}
        |
        |   it_levels = traces.begin()
        |
        |   while it_levels != traces.end():
        |       level = deref(it_levels).first
        |       level_traces = deref(it_levels).second
        |       it_level_traces = level_traces.begin()
        |       result[level] = []
        |       while it_level_traces != level_traces.end():
        |           trace = deref(it_level_traces)
        |           py_trace = [trace[k] for k in range(trace.size())]
        |           result[level].append(py_trace)
        |           inc(it_level_traces)
        |       inc(it_levels)
        |   return result
        |
        |
        |def set_sec_factor(double factor):
        |    _set_sec_factor(factor)
        |
        |ctypedef vector[double]* vec_double_ptr
        |cdef unordered_map[long, vec_double_ptr] _hvalues = unordered_map[long, vec_double_ptr ]()
        |
        |cdef void _hvalue_callback(long *handle, double *h) nogil:
        |    cdef vector[double]* hvalues_c = _hvalues.at(deref(handle))
        |    deref(hvalues_c).push_back(deref(h))
        """
    )



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import glob
import os
import shutil
import subprocess

from setuptools._distutils import ccompiler

from .utils import (
    base_cache_folder,
    create_folder_if_not_exists,
    get_platform,
    run_in_folder,
)

CC = os.environ.get("CC", ccompiler.new_compiler().executables["compiler"][0])
HERE = os.path.dirname(os.path.abspath(__file__))


def compile_if_needed(target_folder):
    lsoda_out = os.path.join(target_folder, "lsoda.a")
    if not os.path.exists(lsoda_out):
        with run_in_folder(target_folder):
            for p in glob.glob(os.path.join(HERE, "lsoda_modified", "*.[ch]")):
                shutil.copy(p, ".")

            c_files = glob.glob("*.c")

            output = subprocess.check_output(
                [
                    CC,
                    "-O3",
                    "-c",
                    "-fPIC",
                    "-DINTEGER_STAR_8=1",
                    "-w",
                    "-Wfatal-errors",
                ]
                + c_files,
                shell=False,
                universal_newlines=True,
            )
            print(output)

            o_files = glob.glob("*.o")
            output = subprocess.check_output(
                [
                    "ar",
                    "rcs",
                    lsoda_out,
                ]
                + o_files,
                shell=False,
                universal_newlines=True,
            )
            print(output)
    return lsoda_out


def install_lsoda_fast_if_needed(lsoda_root_folder=None):
    if lsoda_root_folder is None:
        lsoda_root_folder = os.path.join(base_cache_folder(), "sympy2c", get_platform())

    target_folder = os.path.join(lsoda_root_folder, "lsoda_fast")
    create_folder_if_not_exists(target_folder)
    object_file = compile_if_needed(target_folder)
    return object_file



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import os
import tarfile

import requests

from .utils import run_in_folder

FILE_NAME = "gsl-latest.tar.gz"
URL = "https://ftp.gnu.org/gnu/gsl/" + FILE_NAME


def download_if_needed(download_folder, FILE_NAME=FILE_NAME, URL=URL):
    with run_in_folder(download_folder):
        if not os.path.exists(FILE_NAME):
            print("download source code from", URL)
            with open(FILE_NAME, "wb") as fh:
                fh.write(requests.get(URL).content)


def decompress_if_needed(download_folder, target_folder, FILE_NAME=FILE_NAME):
    # gsl_folder = target_folder
    with run_in_folder(os.path.join(download_folder, target_folder)):
        with tarfile.open(os.path.join(download_folder, FILE_NAME), "r") as fh:
            gsl_folder = os.path.join(target_folder, fh.getnames()[0])
            if not os.path.exists(gsl_folder):
                print("extract", FILE_NAME)
                fh.extractall()
        return gsl_folder


def configure_if_needed(gsl_folder, target):
    with run_in_folder(gsl_folder):
        if not os.path.exists("Makefile"):
            assert (
                os.system("./configure --prefix={target}".format(target=target)) == 0
            ), "running configure failed"


def run_make_if_needed(target_folder):
    with run_in_folder(target_folder):
        if not os.path.exists("./statistics/ttest.o"):
            assert os.system("make") == 0, "running make failed"


def run_make_install_if_needed(gsl_folder, target_folder):
    with run_in_folder(gsl_folder):
        if not all(
            os.path.exists(os.path.join(target_folder, sub_folder))
            for sub_folder in ("lib", "bin", "include")
        ):
            assert os.system("make install") == 0, "running make install failed"


def install_gsl_if_needed(download_folder, target_folder):
    download_if_needed(download_folder)
    gsl_folder = decompress_if_needed(download_folder, target_folder)
    configure_if_needed(gsl_folder, target_folder)
    run_make_if_needed(gsl_folder)
    run_make_install_if_needed(gsl_folder, target_folder)
    return target_folder


if __name__ == "__main__":
    install_gsl_if_needed("/tmp/gsl_download", "/tmp/gsl_installation")



# This file is part of PyCosmo, a multipurpose cosmology calculation tool in Python.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.


import glob
import importlib
import inspect
import os
from pickle import dumps, loads

from sympy2c import (
    ERROR,
    Alias,
    Function,
    Globals,
    Integral,
    Module,
    Symbol,
    Vector,
    symbols,
)


def test_all_classes_with_own_new_method_for_pickling():
    """
    Pickling can fail if classes implement their own __new__ method.
    Here we extract all classes implemented in sympy2c and check if
    they have their own __new__ method implemented within sympy2c.
    For those classes we run tests.
    """

    import sympy2c

    folder = sympy2c.__path__[0]

    setup_functions = {
        "VectorElement": _setup_pickling_vector_element,
        "Vector": _setup_pickling_vector,
        "Symbol": _setup_pickling_symbol,
        "Integral": _setup_pickling_integral,
        "Function": _setup_pickling_function,
        "InternalFunction": None,
        "ERROR": _setup_pickling_error,
    }

    tested = set()

    for p in glob.glob(folder + "/**/*.py", recursive=True):
        # fix path relative to folder above sympy package:
        p = os.path.relpath(p, os.path.dirname(folder))

        # filepath to module notation:
        module_specificer = os.path.splitext(p)[0].replace("/", ".")
        module = importlib.import_module(module_specificer)

        for item in module.__dict__.values():
            if not inspect.isclass(item):
                continue

            if (getattr(item.__new__, "__module__") or "").startswith("sympy2c."):
                item_name = str(item)
                if item_name in tested:
                    continue
                tested.add(item_name)
                if "." in item_name:
                    # extract name from "<class ....."> notation.
                    # some classes derived from sympy.Function don't have this
                    # struture.
                    item_name = item_name.split("'")[1].split(".")[-1]
                if item_name not in setup_functions:
                    assert False, "no test function implemented for {}".format(
                        item_name
                    )
                value = setup_functions[item_name]
                if value is not None:
                    value2 = loads(dumps(value))
                    assert value == value2
                    if hasattr(value, "free_symbols"):
                        assert value.free_symbols == value2.free_symbols


def _setup_pickling_vector():
    return Vector("v", 10)


def _setup_pickling_vector_element():
    return Vector("v", 10)[0]


def _setup_pickling_integral():
    x = Symbol("x")
    ii = Integral(x**2, x, 0, 1)
    return ii


def _setup_pickling_function():
    x = Symbol("x")
    f = Function("f", x * x, x)
    return f


def _setup_pickling_symbol():
    x = Symbol("x")
    return x


def _setup_pickling_error():
    e = ERROR("message")
    return e


def test_pickling_module_and_wrapper(compile_and_load):
    s = Symbol("s")
    t = Symbol("t")

    f = Function("f", s * t, s)
    g = Globals(t)

    m = Module()
    m.add(g)
    m.add(f)

    assert loads(dumps(m)).get_unique_id() == m.get_unique_id()

    w = compile_and_load(m)

    w_back = loads(dumps(w))

    w.globals.t = 2
    w_back.globals.t = 2

    assert w.f(3) == w_back.f(3) == 6


def test_pickling_function_with_alias():
    u, v, w = symbols("u v w")
    y = Alias("y", u, v, w)
    f = Function("y", u, y)

    f_back = loads(dumps(f))
    assert f_back.aliases == f.aliases



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import sympy as sp

from .function import (
    Alias,
    Function,
    MatrixValuedFunctionWrapper,
    VectorValuedFunctionWrapper,
)
from .utils import Hasher, align, concat_generator_results
from .wrapper import WrapperBase


class Ode:
    def __init__(self, name, t, lhs, rhs_expressions, *, compute_jacobian=True):
        assert len(lhs) == len(rhs_expressions)
        self.name = name
        self.t = t
        self.lhs = lhs
        self.rhs_expressions = rhs_expressions
        self.compute_jacobian = compute_jacobian
        self._unique_id = None

    def __str__(self):
        return (
            "<Ode "
            + " ".join(
                map(
                    str,
                    (
                        self.name,
                        self.t,
                        self.lhs,
                        self.rhs_expressions,
                        self.compute_jacobian,
                    ),
                )
            )
            + ">"
        )

    @property
    def wrapper(self):
        return OdeWrapper

    def get_unique_id(self):
        if self._unique_id is None:
            hasher = Hasher()
            hasher.update("name", self.name)
            hasher.update("t", str(self.t))
            hasher.update("lhs", str(self.lhs))
            hasher.update("compute_jacobian", str(self.compute_jacobian))
            self._unique_id = hasher.hexdigest()
        return self._unique_id


class OdeWrapper(WrapperBase):
    def __init__(self, ode, globals_, visitor):
        self.ode = ode
        self.globals_ = globals_
        self.visitor = visitor

        self.name = ode.name
        self.t = ode.t
        self.rhs_expressions = ode.rhs_expressions
        self.lhs = ode.lhs

        self.n = len(self.lhs)

    def get_unique_id(self):
        return self.ode.get_unique_id()

    def setup_code_generation(self):
        y = Alias("y___", *self.lhs)

        self.ydot = Function(
            "{}_dot".format(self.name), self.rhs_expressions, self.t, y
        )

        if self.ode.compute_jacobian:
            matrix_jacobian = sp.Matrix(self.rhs_expressions).jacobian(self.lhs)
            columnwise_vector_jacobian = matrix_jacobian.T[:]
            self.jacobian = Function(
                "{}_jacobian".format(self.name), columnwise_vector_jacobian, self.t, y
            )
        else:
            self.jacobian = None

        self.wrappers = [
            VectorValuedFunctionWrapper(self.ydot, self.globals_, self.visitor, False)
        ]
        if self.jacobian is not None:
            self.wrappers.append(
                MatrixValuedFunctionWrapper(
                    self.jacobian, self.globals_, self.visitor, False
                )
            )

        for w in self.wrappers:
            w.setup_code_generation()

    @concat_generator_results
    def c_header(self):
        for wrapper in self.wrappers:
            yield wrapper.c_header()

        yield (
            "int _solve_{name}("
            "double *, double *, size_t, double *, double *, int, int, int);"
        ).format(name=self.name)

    @concat_generator_results
    def c_code(self, header_file_path):
        for wrapper in self.wrappers:
            yield wrapper.c_code(header_file_path)
        yield align(
            """
        |void _fex_{name}(double t, double *y, double *ydot, void *data)
        |{{
        |    _{name}_dot(t, y, ydot);
        |}}
        """.format(name=self.name)
        )

        if self.jacobian:
            yield align(
                """
            |void _fjac_{name}(double t, double *y, double *jac, void *data)
            |{{
            |    _{name}_jacobian(t, y, jac);
            |}}
            """.format(name=self.name)
            )

        if self.ode.compute_jacobian:
            jac = "use_jacobian ? _fjac_{name} : NULL".format(name=self.name)
        else:
            jac = "NULL"

        yield align(
            """
        |int _solve_{name}(
        |    double *tgrid, double *y, size_t nt, double * atol, double * rtol,
        |    int max_iter, int max_order, int use_jacobian
        |)
        |{{
        |
        |    int istate = 1;
        |    double t;
        |
        |    int iwork1 = 0;
        |    int iwork2 = 0;
        |    int iwork5 = 1;     // report solver switches
        |    int iwork6 = max_iter;
        |    int iwork7 = 0;
        |    int iwork8 = 0;
        |    int iwork9 = max_order;
        |    int iopt = 1;
        |    int jt = 2;    // lsoda only uses jac in case jac != NULL and else uses
        |                   // approximation.
        |    int itol = 4;  // rtol and atol as vector
        |
        |    // use default values:
        |    double rwork1 = 0.0;  // tcrit
        |    double rwork5 = 0.0;  // h0
        |    double rwork6 = 0.0;  // hmax
        |    double rwork7 = 0.0;  // hmin
        |    void * data = 0;
        |
        |    for (size_t i = 1; i < nt; ++i) {{
        |         t = tgrid[i - 1];
        |         memcpy(y + i * {n}, y + (i - 1) * {n}, {n} * sizeof(double));
        |         /* we subtrace 1 because lsoda uses offset 1 */
        |         lsoda(_fex_{name}, {jac}, {n}, y + i * {n} - 1, &t, tgrid[i], itol,
        |               rtol, atol, 1, &istate, iopt, jt, iwork1, iwork2, iwork5,
        |               iwork6, iwork7, iwork8, iwork9, rwork1, rwork5, rwork6,
        |               rwork7, data);
        |         if (istate <= 0)
        |            break;
        |    }}
        |
        |   n_lsoda_terminate();
        |   return istate;
        |
        |}}
    """
        ).format(name=self.name, n=self.n, jac=jac)

    @concat_generator_results
    def cython_code(self, header_file_path):
        for wrapper in self.wrappers:
            yield wrapper.cython_code(header_file_path)

        yield align(
            """
        |cdef extern from "{header_file_path}":
        |
        |    int _solve_{name}(double *tgrid, double *y, size_t nt, double * atol,
        |                      double * rtol, int max_iter, int max_order,
        |                      int use_jacobian)
        |
        |    void _{name}_dot(double t, double *y, double *ydot)
        """
            + (
                ""
                if not self.ode.compute_jacobian
                else """
        |    void _{name}_jacobian(double t, double *y, double *jac)
        |
        """
            )
            + """
        |import warnings
        |
        |def solve_{name}(
        |    np.ndarray[np.double_t, ndim=1] y0,
        |    np.ndarray[np.double_t, ndim=1] tgrid,
        |    rtol, atol, int max_iter=0, int max_order=5, int use_jacobian=1
        |):
        |    '''solve_{name}(y0, tgrid, atol, rtol, max_iter, max_order, use_jacobian)
        |
        |       default value for max_iter is 0 which indicates "use default value
        |       of lsoda"
        |    '''
        |
        |
        |    cdef size_t nt = tgrid.shape[0]
        |    cdef int state
        |    assert y0.shape[0] == {n}, "y0 has wrong shape"
        |    assert not (use_jacobian > 0) or {compute_jacobian}, (
        |          "you must set 'compute_jacobian' to True when "
        |          "creating the ode wrapper.")
        |    assert isinstance(atol, (float, np.ndarray))
        |    if isinstance(atol, float):
        |         atol = atol * np.ones(({n},))
        |    assert atol.shape == ({n},)
        |
        |    assert isinstance(rtol, (float, np.ndarray))
        |    if isinstance(rtol, float):
        |         rtol = rtol * np.ones(({n},))
        |    assert rtol.shape == ({n},)
        |
        |    assert max_iter >= 0, "max_iter must be 0 or positive"
        |    assert 1 <= max_order <= 5, "max_order must be in 1..5"
        |    cdef double[{n} + 1] _atol
        |    cdef double[{n} + 1] _rtol
        |    cdef size_t i
        |    for i in range(1, {n} + 1):
        |         _atol[i] = atol[i - 1]
        |         _rtol[i] = rtol[i - 1]
        |
        |    cdef np.ndarray[np.double_t, ndim=2, mode='c'] result
        |    result = np.zeros((nt, {n}), dtype=float, order="c")
        |    result[0, :] = y0
        |    cdef long started = microtime()
        |    istate = _solve_{name}(&tgrid[0], &result[0, 0], nt, &_atol[0], &_rtol[0],
        |                           max_iter, max_order, use_jacobian)
        |
        |    result_meta = dict(istate=istate,
        |                       num_steps=None,
        |                       num_feval=None,
        |                       num_lu_solver_calls=None,
        |                       last_switch_at=None,
        |                       seconds_needed = (microtime() - started) / 1e6,
        |                       steps=None,
        |                       step_sizes=None)
        |    return result, result_meta
        |
        |def symbols_{name}():
        |    return {symbols}
        |
        |def rhs_{name}(double t, np.ndarray[np.double_t, ndim=1] y):
        |    cdef np.ndarray result = np.zeros(({n},))
        |    cdef double[:] result_view = result
        |    _{name}_dot(t, &y[0], &result_view[0])
        |    return result
        |"""
            + (
                ""
                if not self.ode.compute_jacobian
                else """

        |def jac_{name}(double t, np.ndarray[np.double_t, ndim=1] y):
        |    cdef np.ndarray result = np.zeros(({n}, {n}))
        |    cdef double[:, :] result_view = result
        |    _{name}_jacobian(t, &y[0], &result_view[0, 0])
        |    return result
        |
        |"""
            )
        ).format(
            header_file_path=header_file_path,
            name=self.name,
            n=self.n,
            symbols=[str(sym) for sym in self.lhs],
            compute_jacobian=self.ode.compute_jacobian,
        )

    def determine_required_extra_wrappers(self):
        pass



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import ctypes
import glob
import hashlib
import importlib
import os
import subprocess
import sys
import threading
import time
import warnings
from contextlib import contextmanager
from functools import lru_cache, partial

from .utils import (
    align,
    base_cache_folder,
    create_folder_if_not_exists,
    get_platform,
    sympy2c_cache_folder,
)

HERE = os.path.dirname(os.path.abspath(__file__))


ctypes.pythonapi.PyCapsule_GetPointer.restype = ctypes.c_void_p
ctypes.pythonapi.PyCapsule_GetPointer.argtypes = [ctypes.py_object, ctypes.c_char_p]

ctypes.pythonapi.PyCapsule_GetName.restype = ctypes.c_char_p
ctypes.pythonapi.PyCapsule_GetName.argtypes = [ctypes.py_object]


def compile_if_needed_and_load(
    module_wrapper,
    root_folder,
    lsoda_root_folder,
    gsl_root,
    compilation_flags,
    cache_folder=None,
):
    from .build_f2c import install_f2c_if_needed
    from .build_gsl import install_gsl_if_needed
    from .build_lsoda import install_lsoda_if_needed
    from .build_lsoda_fast import install_lsoda_fast_if_needed

    if cache_folder is None:
        cache_folder = base_cache_folder()

    if gsl_root is None:
        gsl_root = os.path.join(cache_folder, "sympy2c", get_platform(), "gsl")

    if not os.path.exists(gsl_root) or (
        not os.path.exists(os.path.join(gsl_root, "include"))
        or not os.path.exists(os.path.join(gsl_root, "lib"))
    ):
        download_folder = os.path.join(cache_folder, "sympy2c", "gsl_source")
        install_gsl_if_needed(download_folder, gsl_root)

    bf = build_folder(module_wrapper, compilation_flags, root_folder)

    libf2c_path = install_f2c_if_needed(cache_folder)
    lsoda_object_file = install_lsoda_if_needed(lsoda_root_folder)
    lsoda_fast_static_lib = install_lsoda_fast_if_needed(lsoda_root_folder)

    generate_files_and_compile_if_needed(
        wrapper_id(module_wrapper, compilation_flags),
        bf,
        module_wrapper,
        gsl_root,
        lsoda_object_file,
        lsoda_fast_static_lib,
        libf2c_path,
        compilation_flags,
    )
    module = LoadedModule(bf)
    return module


def generate_files_and_compile_if_needed(
    wrapper_id,
    build_folder,
    module_wrapper,
    gsl_root,
    lsoda_object_file,
    lsoda_fast_static_lib,
    libf2c_path,
    compilation_flags,
):
    from .integral import IntegralFunctionWrapper

    IntegralFunctionWrapper.reset()

    if glob.glob(
        os.path.join(build_folder, "_wrapper_{}.so".format(wrapper_id), "setup.py")
    ):
        print("wrapper already exists at", build_folder)
        return build_folder

    create_folder_if_not_exists(build_folder)

    build = os.environ.get("REBUILD_WRAPPER") or not glob.glob(
        os.path.join(build_folder, "_wrapper_{}.pyx".format(wrapper_id))
    )
    if build:
        print("create source files at", build_folder)
        create_source_files(
            build_folder,
            module_wrapper,
            gsl_root,
            lsoda_object_file,
            lsoda_fast_static_lib,
            libf2c_path,
            wrapper_id,
            compilation_flags,
        )
    else:
        print("source files already exist at", build_folder)
    compile_files(build_folder)


@lru_cache()
def warn_new_traces(unique_id):
    from .ode_fast import read_new_traces

    if read_new_traces(unique_id):
        warnings.warn("there are new permuations pending, you might want to recompile")


class LoadedModule:
    """wraps compiled module to support pickling"""

    def __init__(self, folder):
        self._folder = os.path.abspath(folder)
        self._wrapper_id = os.path.basename(folder)
        self._module_name = "_wrapper_" + self._wrapper_id
        self._load()

    def _load(self, state=None):
        if self._module_name in sys.modules:
            del sys.modules[self._module_name]
        try:
            before = os.getcwd()
            os.chdir(self._folder)
            sys.path.insert(0, self._folder)
            self._module = importlib.import_module(self._module_name)
            if state is not None:
                self._module.__setstate__(state)
        except FileNotFoundError:
            raise ImportError()
        finally:
            sys.path.pop(0)
            os.chdir(before)
        for unique_id in self._module.get_fast_ode_unique_ids().values():
            warn_new_traces(unique_id)

    def reload(self):
        self._load()

    def __getattr__(self, name):
        if name.startswith("solve_fast"):
            solver = getattr(self._module, name)
            return partial(solver, self)
        if name in self._module._api:
            _fun = setup_function(name, self._module)
            setattr(self._module, name, _fun)
            return _fun

        return getattr(self._module, name)

    def __getstate__(self):
        return self._folder, self._module_name, self._module.__getstate__()

    def __setstate__(self, data):
        try:
            self._folder, self._module_name, state = data
            self._load(state)
        except Exception:
            import traceback

            traceback.print_exc()

    def __dir__(self):
        return self._module.__dir__()


@lru_cache
def setup_function(name, module):
    import numba  # noqa: F401
    import numpy as np  # noqa: F401
    from numba import float64, njit  # noqa: F401

    res_spec, arg_types, args, vector_flags, numba_signature = module._api[name]

    capsule = module.__pyx_capi__[f"_api_{name}"]

    fun_ptr = ctypes.pythonapi.PyCapsule_GetPointer(
        capsule, ctypes.pythonapi.PyCapsule_GetName(capsule)
    )
    res_type, res_shape = res_spec
    fun = ctypes.CFUNCTYPE(
        ctypes.c_int,
        *(arg_types + [res_type]),
    )(fun_ptr)

    ctype_args = ", ".join(
        [
            f"{arg}" + (".ctypes" if is_vector else "")
            for (is_vector, arg) in zip(vector_flags, args)
        ]
        + ["__result.ctypes"]
    )
    arg_decl = ", ".join(args)
    numba_decorator = (
        f"@numba.jit([{numba_signature}], nopython=True)"
        if not vector_flags or any(vector_flags)
        else f"@numba.vectorize([{numba_signature}], nopython=True)"
    )
    code = align(
        f"""
    |import numpy as np
    |{numba_decorator}
    |def {name}({arg_decl}):
    |    __result = np.zeros({res_shape if res_shape is not None else (1,)})
    |
    |    retcode = fun({ctype_args})
    |    if retcode == 1:
    |        raise ArithmeticError("calling {name} did not succeed, "
    |                 "use get_last_error() function for details.")
    |    if retcode == 2:
    |        print("calling {name} issued a warning, "
    |              "use get_last_warning() function for details.")
    |    return {"__result[0]" if res_shape is None else "__result"}
    """
    )
    lo = dict(locals())
    gl = dict(globals())
    gl["fun"] = fun
    gl["np"] = np
    exec(code, gl, lo)
    return lo[name]


# for backwards compatibility:
load_module = LoadedModule


def create_source_files(
    build_folder,
    module_wrapper,
    gsl_root,
    lsoda_object_file,
    lsoda_fast_static_lib,
    libf2c_path,
    wrapper_id,
    compilation_flags,
):
    j = partial(os.path.join, build_folder)

    print("setup code generation")
    module_wrapper.setup_code_generation()

    with open(j("functions.cpp"), "w") as fh:
        print(module_wrapper.c_code("functions.hpp"), file=fh)

    with open(j("functions.hpp"), "w") as fh:
        print(module_wrapper.c_header(), file=fh)

    with open(
        j("_wrapper_{wrapper_id}.pyx".format(wrapper_id=wrapper_id)),
        "w",
    ) as fh:
        print(module_wrapper.cython_code("functions.hpp"), file=fh)

    if not compilation_flags:
        compilation_flags = ["-O3"]

    setup_py_content = align(
        """
    |from distutils.core import setup
    |from distutils.extension import Extension
    |from distutils.sysconfig import get_config_vars
    |
    |from Cython.Build import cythonize
    |import numpy as np
    |import glob
    |import os
    |import sys
    |
    |get_config_vars()["CC"] = os.environ.get("CC", get_config_vars()["CC"])
    |get_config_vars()["CXX"] = os.environ.get("CXX", get_config_vars()["CXX"])
    |
    |sourcefiles = ['_wrapper_{wrapper_id}.pyx', 'functions.cpp']
    |
    |if (os.environ.get("CC") == "clang" or get_config_vars()['CC'] == 'clang'
    |    or sys.platform == "darwin"):
    |        libf2c = ['{libf2c}']
    |        link_flags = []
    |else:
    |        libf2c = ['-Wl,--whole-archive', '{libf2c}', '-Wl,--no-whole-archive']
    |        link_flags = ["-Wl,--allow-multiple-definition"]
    |
    |extensions = [Extension("_wrapper_{wrapper_id}",
    |                sourcefiles,
    |                define_macros = [('HAVE_INLINE', '1'),
    |                                 ('CYTHON_EXTERN_C', 'extern "C"')],
    |                include_dirs=['{gsl_root}/include', np.get_include(),],
    |                library_dirs=['{gsl_root}/lib'],
    |                extra_compile_args = {compilation_flags} +
    |                          ["-std=c++11",
    |                           "-fPIC",
    |                           "-DINTEGER_STAR_8=1",
    |                           "-pipe",
    |                           "-Wno-unused-variable",
    |                           "-fno-var-tracking",
    |                           "-Wno-unused-but-set-variable",
    |                           ],
    |                extra_link_args = ["-fPIC",
    |                                   "-u s_stop",
    |                                   ] + link_flags,
    |                extra_objects = (['{gsl_root}/lib/libgsl.a', '{lsoda_object_file}',
    |                                  '{lsoda_fast_static_lib}']
    |                                 + libf2c
    |                                 )
    |                )
    |              ]
    |
    |setup(
    |   ext_modules=cythonize(extensions, language="c++")
    |)
    """
    ).format(
        gsl_root=gsl_root,
        lsoda_object_file=lsoda_object_file,
        lsoda_fast_static_lib=lsoda_fast_static_lib,
        libf2c=libf2c_path,
        wrapper_id=wrapper_id,
        compilation_flags=compilation_flags,
    )

    with open(j("setup.py"), "w") as fh:
        print(setup_py_content, file=fh)


def compile_files(folder):
    if not os.environ.get("REBUILD_WRAPPER"):
        try:
            LoadedModule(folder)
            return
        except ImportError:
            pass
    current_folder = os.getcwd()
    try:
        os.chdir(folder)
        print()
        print("compile python extension module")
        s = time.time()
        run_command("python setup.py build_ext --inplace")
        print(
            "compiling the module needed {:.1f} minutes".format((time.time() - s) / 60)
        )
    finally:
        os.chdir(current_folder)


def run_command(cmd):
    print("$", cmd)
    with monitor():
        proc = subprocess.run(
            cmd,
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
            shell=True,
            universal_newlines=True,
        )
    if proc.returncode:
        print(proc.stdout)
        raise OSError("compilation failed")


@contextmanager
def monitor():
    t = PrintAliveThread()
    t.start()
    try:
        yield
    finally:
        t.running = False
        time.sleep(0.05)
        print()
        t.join()


class PrintAliveThread(threading.Thread):
    def __init__(self):
        super().__init__()
        self.running = True

    def run(self):
        started = time.time()
        while self.running:
            while time.time() < started + 1.0:
                if not self.running:
                    return
                time.sleep(0.05)
            started = time.time()
            print(".", end="")
            sys.stdout.flush()


def build_folder(module_wrapper, compilation_flags, root_folder=None):
    if root_folder is None:
        root_folder = sympy2c_cache_folder()
    return os.path.join(
        root_folder,
        wrapper_id(
            module_wrapper,
            compilation_flags,
        ),
    )


def wrapper_id(module_wrapper, *extra):
    extra_str = ("".join(str(e) for e in extra)).encode("ascii")
    return "{}_{}".format(
        module_wrapper.get_unique_id(), hashlib.md5(extra_str).hexdigest()[:5]
    )



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
import inspect
from textwrap import dedent
from types import FunctionType

from .wrapper import WrapperBase


class PythonFunction:
    def __init__(self, function):
        assert isinstance(function, FunctionType)
        self.function_src = inspect.getsource(function)
        self.name = function.__name__
        self._unique_id = None

    def get_unique_id(self):
        if self._unique_id is None:
            md5 = hashlib.md5()
            md5.update(self.function_src.encode("utf-8"))
            self._unique_id = md5.hexdigest()
        return self._unique_id

    def wrapper(self):
        return PythonFunctionWrapper(self.name, self.function_src)


class PythonFunctionWrapper(WrapperBase):
    def __init__(self, name, function_src):
        self.name = name
        self.function_src = function_src
        self._unique_id = None

    def c_header(self):
        return ""

    def c_code(self, header_file_path):
        return ""

    def cython_code(self, header_file_path):
        return dedent(self.function_src)

    def determine_required_extra_wrappers(self):
        pass

    def setup_code_generation(self):
        pass

    def get_unique_id(self):
        if self._unique_id is None:
            md5 = hashlib.md5()
            md5.update(self.function_src.encode("utf-8"))
            self._unique_id = md5.hexdigest()
        return self._unique_id



# This file is part of PyCosmo, a multipurpose cosmology calculation tool in Python.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

from sympy import Abs, sqrt

from sympy2c import Function, Module, Symbol
from sympy2c.expressions import Max, Min


def test_min(tmpdir, compile_and_load):
    a = Symbol("a")
    b = Symbol("b")

    min_f = Function("f", Min(a + b, a * b), a, b)
    max_f = Function("g", Max(a + b, a * b), a, b)

    m = Module()
    m.add(min_f)
    m.add(max_f)

    w = compile_and_load(m)

    assert w.f(1.0, 2.0) == 2.0
    assert w.f(2.0, 1.0) == 2.0

    assert w.g(1.0, 2.0) == 3.0
    assert w.g(2.0, 1.0) == 3.0


def test_abs(tmpdir, compile_and_load):
    a = Symbol("a")

    f1 = Function("f1", sqrt(a**2), a)
    f2 = Function("f2", Abs(a), a)

    m = Module()
    m.add(f1)
    m.add(f2)

    w = compile_and_load(m)

    assert w.f1(1) == 1.0
    assert w.f1(-1) == 1.0

    assert w.f2(1) == 1.0
    assert w.f2(-1) == 1.0


def test_if_symbols_are_real():
    # sympy2c symbols represent real numbers which supports
    # simplifications:
    a = Symbol("a")
    assert sqrt(a * a) == Abs(a)

    # default Symbols are not real, so simplification does not work:
    import sympy

    a = sympy.Symbol("a")
    assert sqrt(a * a) != Abs(a)



# This file is part of sympy2c.
#
# Copyright (C) 2013-2022 ETH Zurich, Institute for Particle and Astrophysics and SIS
# ID.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

from .utils import align


def create_interface_onary():
    return align(
        """
        |class OneAryFunction {
        |
        |    public:
        |        virtual double operator()(double) = 0;
        |        virtual double operator()(double, void *) = 0;
        |};
        """
    )


def create_curry_class(n_fixed_args):
    """creates code to implement currying of a function.  Given a
    function with n_fixed_args + 1 double arguments, the first
    n_fixed_args are frozen, such that the resulting object behaves like
    a function with one input argument of type double"""

    class_name = "Curry_{}".format(n_fixed_args)

    function_args_decl = ", ".join("double" for i in range(n_fixed_args + 1))

    default_args_decl = "\n    ".join(
        "double _v{i};".format(i=i) for i in range(n_fixed_args)
    )

    cons_args_decl = ", ".join(
        [
            "void (*f)({function_args_decl}, double *)".format(
                function_args_decl=function_args_decl
            )
        ]
        + ["double v{i}".format(i=i) for i in range(n_fixed_args)]
    )

    attributes_init = ", ".join(
        ["_f(f)"] + ["_v{i}(v{i})".format(i=i) for i in range(n_fixed_args)]
    )

    fixed_args = ", ".join(["_v{i}".format(i=i) for i in range(n_fixed_args)] + ["x"])

    code = align(
        """
        |class {class_name}: public OneAryFunction {{
        |
        |      void (*_f)({function_args_decl}, double *);
        |      {default_args_decl}
        |
        |      public:
        |           {class_name}({cons_args_decl}): {attributes_init} {{}};
        |
        |      double operator()(double x) {{
        |           double __result;
        |           _f({fixed_args},  &__result);
        |           return __result;
        |      }}
        |
        |      // for use as a gsl_function:
        |      double operator()(double x, void *) {{
        |           return this->operator()(x);
        |      }}
        |}};"""
    ).format(
        class_name=class_name,
        function_args_decl=function_args_decl,
        default_args_decl=default_args_decl,
        cons_args_decl=cons_args_decl,
        attributes_init=attributes_init,
        fixed_args=fixed_args,
    )

    return class_name, code
